# Copyright 2020 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_config import cfg

from futurist import periodics
from neutron_lib.api.definitions import external_net as extnet_apidef
from neutron_lib.api.definitions import floating_ip_port_forwarding as pf_def
from neutron_lib.api.definitions import provider_net as provnet_apidef
from neutron_lib import constants as n_const
from neutron_lib import context as n_context
from neutron_lib.exceptions import l3 as lib_l3_exc

from neutron.common.ovn import constants as ovn_const
from neutron.common.ovn import utils
from neutron.conf.plugins.ml2.drivers.ovn import ovn_conf as ovn_config
from neutron.db import ovn_revision_numbers_db as db_rev
from neutron.plugins.ml2.drivers.ovn.mech_driver.ovsdb import maintenance
from neutron.tests.functional import base
from neutron.tests.functional.services.logapi.drivers.ovn \
    import test_driver as test_log_driver

from neutron.tests.unit.api import test_extensions
from neutron.tests.unit.extensions import test_extraroute

CFG_NEW_BURST = 50
CFG_NEW_RATE = 150


class _TestMaintenanceHelper(base.TestOVNFunctionalBase):
    """A helper class to keep the code more organized."""

    def setUp(self):
        super(_TestMaintenanceHelper, self).setUp()
        self._ovn_client = self.mech_driver._ovn_client
        self._l3_ovn_client = self.l3_plugin._ovn_client
        ext_mgr = test_extraroute.ExtraRouteTestExtensionManager()
        self.ext_api = test_extensions.setup_extensions_middleware(ext_mgr)
        self.maint = maintenance.DBInconsistenciesPeriodics(self._ovn_client)
        # Release the unneeded lock
        self.maint._idl.set_lock(None)
        self.context = n_context.get_admin_context()
        # Always verify inconsistencies for all objects.
        db_rev.INCONSISTENCIES_OLDER_THAN = -1

    def _find_network_row_by_name(self, name):
        for row in self.nb_api._tables['Logical_Switch'].rows.values():
            if (row.external_ids.get(
                    ovn_const.OVN_NETWORK_NAME_EXT_ID_KEY) == name):
                return row

    def _create_network(self, name, external=False, provider=None):
        data = {'network': {'name': name,
                            extnet_apidef.EXTERNAL: external}}
        if provider:
            data['network'][provnet_apidef.NETWORK_TYPE] = 'flat'
            data['network'][provnet_apidef.PHYSICAL_NETWORK] = provider
        req = self.new_create_request('networks', data, self.fmt,
                                      as_admin=True)
        res = req.get_response(self.api)
        return self.deserialize(self.fmt, res)['network']

    def _update_network_name(self, net_id, new_name):
        data = {'network': {'name': new_name}}
        req = self.new_update_request('networks', data, net_id, self.fmt)
        res = req.get_response(self.api)
        return self.deserialize(self.fmt, res)['network']

    def _create_port(self, name, net_id, security_groups=None,
                     device_owner=None, vnic_type=None):
        data = {'port': {'name': name,
                         'network_id': net_id}}

        if security_groups is not None:
            data['port']['security_groups'] = security_groups

        if device_owner is not None:
            data['port']['device_owner'] = device_owner

        if vnic_type is not None:
            data['port']['binding:vnic_type'] = vnic_type

        req = self.new_create_request('ports', data, self.fmt)
        res = req.get_response(self.api)
        return self.deserialize(self.fmt, res)['port']

    def _update_port_name(self, port_id, new_name):
        data = {'port': {'name': new_name}}
        req = self.new_update_request('ports', data, port_id, self.fmt)
        res = req.get_response(self.api)
        return self.deserialize(self.fmt, res)['port']

    def _find_port_row_by_name(self, name):
        for row in self.nb_api._tables['Logical_Switch_Port'].rows.values():
            if (row.external_ids.get(
                    ovn_const.OVN_PORT_NAME_EXT_ID_KEY) == name):
                return row

    def _set_global_dhcp_opts(self, ip_version, opts):
        opt_string = ','.join(['{0}:{1}'.format(key, value)
                               for key, value
                               in opts.items()])
        if ip_version == n_const.IP_VERSION_6:
            ovn_config.cfg.CONF.set_override('ovn_dhcp6_global_options',
                                             opt_string,
                                             group='ovn')
        if ip_version == n_const.IP_VERSION_4:
            ovn_config.cfg.CONF.set_override('ovn_dhcp4_global_options',
                                             opt_string,
                                             group='ovn')

    def _unset_global_dhcp_opts(self, ip_version):
        if ip_version == n_const.IP_VERSION_6:
            ovn_config.cfg.CONF.clear_override('ovn_dhcp6_global_options',
                                               group='ovn')
        if ip_version == n_const.IP_VERSION_4:
            ovn_config.cfg.CONF.clear_override('ovn_dhcp4_global_options',
                                               group='ovn')

    def _create_subnet(self, name, net_id, ip_version=n_const.IP_VERSION_4,
                       **kwargs):
        if ip_version == n_const.IP_VERSION_4:
            cidr = '10.0.0.0/24'
        else:
            cidr = '2001:db8::/64'
        data = {'subnet': {'name': name,
                           'network_id': net_id,
                           'ip_version': ip_version,
                           'cidr': cidr,
                           'enable_dhcp': True}}
        data['subnet'].update(kwargs)
        req = self.new_create_request('subnets', data, self.fmt)
        res = req.get_response(self.api)
        return self.deserialize(self.fmt, res)['subnet']

    def _update_subnet_enable_dhcp(self, subnet_id, value):
        data = {'subnet': {'enable_dhcp': value}}
        req = self.new_update_request('subnets', data, subnet_id, self.fmt)
        res = req.get_response(self.api)
        return self.deserialize(self.fmt, res)['subnet']

    def _find_subnet_row_by_id(self, subnet_id):
        for row in self.nb_api._tables['DHCP_Options'].rows.values():
            if (row.external_ids.get('subnet_id') == subnet_id and
               not row.external_ids.get('port_id')):
                return row

    def _create_router(self, name, external_gateway_info=None):
        data = {'router': {'name': name}}
        as_admin = False
        if external_gateway_info is not None:
            data['router']['external_gateway_info'] = external_gateway_info
            as_admin = bool(external_gateway_info.get('enable_snat'))
        req = self.new_create_request('routers', data, self.fmt,
                                      as_admin=as_admin)
        res = req.get_response(self.api)
        return self.deserialize(self.fmt, res)['router']

    def _update_router_name(self, net_id, new_name):
        data = {'router': {'name': new_name}}
        req = self.new_update_request('routers', data, net_id, self.fmt)
        res = req.get_response(self.api)
        return self.deserialize(self.fmt, res)['router']

    def _find_router_row_by_name(self, name):
        for row in self.nb_api._tables['Logical_Router'].rows.values():
            if (row.external_ids.get(
                    ovn_const.OVN_ROUTER_NAME_EXT_ID_KEY) == name):
                return row

    def _create_security_group(self):
        data = {'security_group': {'name': 'sgtest',
                                   'description': 'SpongeBob Rocks!'}}
        req = self.new_create_request('security-groups', data, self.fmt)
        res = req.get_response(self.api)
        return self.deserialize(self.fmt, res)['security_group']

    def _find_security_group_row_by_id(self, sg_id):
        return self.nb_api.lookup(
            'Port_Group', utils.ovn_port_group_name(sg_id), default=None)

    def _create_security_group_rule(self, sg_id):
        data = {'security_group_rule': {'security_group_id': sg_id,
                                        'direction': 'ingress',
                                        'protocol': n_const.PROTO_NAME_TCP,
                                        'ethertype': n_const.IPv4,
                                        'port_range_min': 22,
                                        'port_range_max': 22}}
        req = self.new_create_request('security-group-rules', data, self.fmt)
        res = req.get_response(self.api)
        return self.deserialize(self.fmt, res)['security_group_rule']

    def _find_security_group_rule_row_by_id(self, sgr_id):
        for row in self.nb_api._tables['ACL'].rows.values():
            if (row.external_ids.get(
                    ovn_const.OVN_SG_RULE_EXT_ID_KEY) == sgr_id):
                return row

    def _process_router_interface(self, action, router_id, subnet_id):
        req = self.new_action_request(
            'routers', {'subnet_id': subnet_id}, router_id,
            '%s_router_interface' % action)
        res = req.get_response(self.api)
        return self.deserialize(self.fmt, res)

    def _add_router_interface(self, router_id, subnet_id):
        return self._process_router_interface('add', router_id, subnet_id)

    def _remove_router_interface(self, router_id, subnet_id):
        return self._process_router_interface('remove', router_id, subnet_id)

    def _find_router_port_row_by_port_id(self, port_id):
        for row in self.nb_api._tables['Logical_Router_Port'].rows.values():
            if row.name == utils.ovn_lrouter_port_name(port_id):
                return row

    def _find_nat_rule(self, router_id, external_ip, logical_ip=None,
                       nat_type='dnat_and_snat'):
        rules = self.nb_api.get_lrouter_nat_rules(utils.ovn_name(router_id))
        return next((r for r in rules
                     if r['type'] == nat_type and
                     r['external_ip'] == external_ip and
                     (not logical_ip or r['logical_ip'] == logical_ip)),
                    None)

    def _find_pf_lb(self, router_id, fip_id=None):
        lbs = self.nb_api.get_router_floatingip_lbs(utils.ovn_name(router_id))
        return [lb for lb in lbs
                if (not fip_id or
                    fip_id == lb.external_ids[ovn_const.OVN_FIP_EXT_ID_KEY])]


class TestMaintenance(_TestMaintenanceHelper):

    def test_network(self):
        net_name = 'networktest'
        with mock.patch.object(self._ovn_client, 'create_network'):
            neutron_obj = self._create_network(net_name)

        # Assert the network doesn't exist in OVN
        self.assertIsNone(self._find_network_row_by_name(net_name))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the network was now created
        ovn_obj = self._find_network_row_by_name(net_name)
        self.assertIsNotNone(ovn_obj)
        self.assertEqual(
            neutron_obj['revision_number'],
            int(ovn_obj.external_ids[ovn_const.OVN_REV_NUM_EXT_ID_KEY]))

        # > Update

        new_obj_name = 'networktest_updated'
        with mock.patch.object(self._ovn_client, 'update_network'):
            new_neutron_obj = self._update_network_name(neutron_obj['id'],
                                                        new_obj_name)

        # Assert the revision numbers are out-of-sync
        ovn_obj = self._find_network_row_by_name(net_name)
        self.assertNotEqual(
            new_neutron_obj['revision_number'],
            int(ovn_obj.external_ids[ovn_const.OVN_REV_NUM_EXT_ID_KEY]))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the old name doesn't exist anymore in the OVNDB
        self.assertIsNone(self._find_network_row_by_name(net_name))

        # Assert the network is now in sync
        ovn_obj = self._find_network_row_by_name(new_obj_name)
        self.assertEqual(
            new_neutron_obj['revision_number'],
            int(ovn_obj.external_ids[ovn_const.OVN_REV_NUM_EXT_ID_KEY]))

        # > Delete

        with mock.patch.object(self._ovn_client, 'delete_network'):
            self._delete('networks', new_neutron_obj['id'])

        # Assert the network still exists in OVNDB
        self.assertIsNotNone(self._find_network_row_by_name(new_obj_name))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the network is now deleted from OVNDB
        self.assertIsNone(self._find_network_row_by_name(new_obj_name))

        # Assert the revision number no longer exists
        self.assertIsNone(db_rev.get_revision_row(
            self.context,
            new_neutron_obj['id']))

    def test_port(self):
        obj_name = 'porttest'
        neutron_net = self._create_network('network1')

        with mock.patch.object(self._ovn_client, 'create_port'):
            neutron_obj = self._create_port(obj_name, neutron_net['id'])

        # Assert the port doesn't exist in OVN
        self.assertIsNone(self._find_port_row_by_name(obj_name))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the port was now created
        ovn_obj = self._find_port_row_by_name(obj_name)
        self.assertIsNotNone(ovn_obj)
        self.assertEqual(
            neutron_obj['revision_number'],
            int(ovn_obj.external_ids[ovn_const.OVN_REV_NUM_EXT_ID_KEY]))

        # > Update

        new_obj_name = 'porttest_updated'
        with mock.patch.object(self._ovn_client, 'update_port'):
            new_neutron_obj = self._update_port_name(neutron_obj['id'],
                                                     new_obj_name)

        # Assert the revision numbers are out-of-sync
        ovn_obj = self._find_port_row_by_name(obj_name)
        self.assertNotEqual(
            new_neutron_obj['revision_number'],
            int(ovn_obj.external_ids[ovn_const.OVN_REV_NUM_EXT_ID_KEY]))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the old name doesn't exist anymore in the OVNDB
        self.assertIsNone(self._find_port_row_by_name(obj_name))

        # Assert the port is now in sync. Note that for ports we are
        # fetching it again from the Neutron database prior to comparison
        # because of the monitor code that can update the ports again upon
        # changes to it.
        ovn_obj = self._find_port_row_by_name(new_obj_name)
        new_neutron_obj = self._ovn_client._plugin.get_port(
            self.context, neutron_obj['id'])
        self.assertEqual(
            new_neutron_obj['revision_number'],
            int(ovn_obj.external_ids[ovn_const.OVN_REV_NUM_EXT_ID_KEY]))

        # > Delete

        with mock.patch.object(self._ovn_client, 'delete_port'):
            self._delete('ports', new_neutron_obj['id'])

        # Assert the port still exists in OVNDB
        self.assertIsNotNone(self._find_port_row_by_name(new_obj_name))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the port is now deleted from OVNDB
        self.assertIsNone(self._find_port_row_by_name(new_obj_name))

        # Assert the revision number no longer exists
        self.assertIsNone(db_rev.get_revision_row(
            self.context,
            neutron_obj['id']))

    def test_subnet_global_dhcp4_opts(self):
        obj_name = 'globaltestsubnet'
        options = {'ntp_server': '1.2.3.4'}
        neutron_net = self._create_network('network1')

        # Create a subnet without global options
        neutron_sub = self._create_subnet(obj_name, neutron_net['id'])

        # Assert that the option is not set
        ovn_obj = self._find_subnet_row_by_id(neutron_sub['id'])
        self.assertIsNone(ovn_obj.options.get('ntp_server', None))

        # Set some global DHCP Options
        self._set_global_dhcp_opts(ip_version=n_const.IP_VERSION_4,
                                   opts=options)

        # Run the maintenance task to add the new options
        self.assertRaises(periodics.NeverAgain,
                          self.maint.check_global_dhcp_opts)

        # Assert that the option was added
        ovn_obj = self._find_subnet_row_by_id(neutron_sub['id'])
        self.assertEqual(
            '1.2.3.4',
            ovn_obj.options.get('ntp_server', None))

        # Change the global option
        new_options = {'ntp_server': '4.3.2.1'}
        self._set_global_dhcp_opts(ip_version=n_const.IP_VERSION_4,
                                   opts=new_options)

        # Run the maintenance task to update the options
        self.assertRaises(periodics.NeverAgain,
                          self.maint.check_global_dhcp_opts)

        # Assert that the option was changed
        ovn_obj = self._find_subnet_row_by_id(neutron_sub['id'])
        self.assertEqual(
            '4.3.2.1',
            ovn_obj.options.get('ntp_server', None))

        # Change the global option to null
        new_options = {'ntp_server': ''}
        self._set_global_dhcp_opts(ip_version=n_const.IP_VERSION_4,
                                   opts=new_options)

        # Run the maintenance task to update the options
        self.assertRaises(periodics.NeverAgain,
                          self.maint.check_global_dhcp_opts)

        # Assert that the option was removed
        ovn_obj = self._find_subnet_row_by_id(neutron_sub['id'])
        self.assertIsNone(ovn_obj.options.get('ntp_server', None))

    def test_subnet_global_dhcp6_opts(self):
        obj_name = 'globaltestsubnet'
        options = {'ntp_server': '1.2.3.4'}
        neutron_net = self._create_network('network1')

        # Create a subnet without global options
        neutron_sub = self._create_subnet(obj_name, neutron_net['id'],
                                          n_const.IP_VERSION_6)

        # Assert that the option is not set
        ovn_obj = self._find_subnet_row_by_id(neutron_sub['id'])
        self.assertIsNone(ovn_obj.options.get('ntp_server', None))

        # Set some global DHCP Options
        self._set_global_dhcp_opts(ip_version=n_const.IP_VERSION_6,
                                   opts=options)

        # Run the maintenance task to add the new options
        self.assertRaises(periodics.NeverAgain,
                          self.maint.check_global_dhcp_opts)

        # Assert that the option was added
        ovn_obj = self._find_subnet_row_by_id(neutron_sub['id'])
        self.assertEqual(
            '1.2.3.4',
            ovn_obj.options.get('ntp_server', None))

        # Change the global option
        new_options = {'ntp_server': '4.3.2.1'}
        self._set_global_dhcp_opts(ip_version=n_const.IP_VERSION_6,
                                   opts=new_options)

        # Run the maintenance task to update the options
        self.assertRaises(periodics.NeverAgain,
                          self.maint.check_global_dhcp_opts)

        # Assert that the option was changed
        ovn_obj = self._find_subnet_row_by_id(neutron_sub['id'])
        self.assertEqual(
            '4.3.2.1',
            ovn_obj.options.get('ntp_server', None))

        # Change the global option to null
        new_options = {'ntp_server': ''}
        self._set_global_dhcp_opts(ip_version=n_const.IP_VERSION_6,
                                   opts=new_options)

        # Run the maintenance task to update the options
        self.assertRaises(periodics.NeverAgain,
                          self.maint.check_global_dhcp_opts)

        # Assert that the option was removed
        ovn_obj = self._find_subnet_row_by_id(neutron_sub['id'])
        self.assertIsNone(ovn_obj.options.get('ntp_server', None))

    def test_subnet(self):
        obj_name = 'subnettest'
        neutron_net = self._create_network('network1')

        with mock.patch.object(self._ovn_client, 'create_subnet'):
            neutron_obj = self._create_subnet(obj_name, neutron_net['id'])

        # Assert the subnet doesn't exist in OVN
        self.assertIsNone(self._find_subnet_row_by_id(neutron_obj['id']))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the subnet was now created
        ovn_obj = self._find_subnet_row_by_id(neutron_obj['id'])
        self.assertIsNotNone(ovn_obj)
        self.assertEqual(
            neutron_obj['revision_number'],
            int(ovn_obj.external_ids[ovn_const.OVN_REV_NUM_EXT_ID_KEY]))

        # > Update

        with mock.patch.object(self._ovn_client, 'update_subnet'):
            neutron_obj = self._update_subnet_enable_dhcp(
                neutron_obj['id'], False)

        # Assert the revision numbers are out-of-sync
        ovn_obj = self._find_subnet_row_by_id(neutron_obj['id'])
        self.assertNotEqual(
            neutron_obj['revision_number'],
            int(ovn_obj.external_ids[ovn_const.OVN_REV_NUM_EXT_ID_KEY]))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the old name doesn't exist anymore in the OVNDB. When
        # the subnet's enable_dhcp's is set to False, OVN will remove the
        # DHCP_Options entry related to that subnet.
        self.assertIsNone(self._find_subnet_row_by_id(neutron_obj['id']))

        # Re-enable the DHCP for the subnet and check if the maintenance
        # thread will re-create it in OVN
        with mock.patch.object(self._ovn_client, 'update_subnet'):
            neutron_obj = self._update_subnet_enable_dhcp(
                neutron_obj['id'], True)

        # Assert the DHCP_Options still doesn't exist in OVNDB
        self.assertIsNone(self._find_subnet_row_by_id(neutron_obj['id']))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the subnet is now in sync
        ovn_obj = self._find_subnet_row_by_id(neutron_obj['id'])
        self.assertEqual(
            neutron_obj['revision_number'],
            int(ovn_obj.external_ids[ovn_const.OVN_REV_NUM_EXT_ID_KEY]))

        # > Delete

        with mock.patch.object(self._ovn_client, 'delete_subnet'):
            self._delete('subnets', neutron_obj['id'])

        # Assert the subnet still exists in OVNDB
        self.assertIsNotNone(self._find_subnet_row_by_id(neutron_obj['id']))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the subnet is now deleted from OVNDB
        self.assertIsNone(self._find_subnet_row_by_id(neutron_obj['id']))

        # Assert the revision number no longer exists
        self.assertIsNone(db_rev.get_revision_row(
            self.context,
            neutron_obj['id']))

    def test_router(self):
        obj_name = 'routertest'

        with mock.patch.object(self._l3_ovn_client, 'create_router'):
            neutron_obj = self._create_router(obj_name)

        # Assert the router doesn't exist in OVN
        self.assertIsNone(self._find_router_row_by_name(obj_name))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the router was now created
        ovn_obj = self._find_router_row_by_name(obj_name)
        self.assertIsNotNone(ovn_obj)
        self.assertEqual(
            neutron_obj['revision_number'],
            int(ovn_obj.external_ids[ovn_const.OVN_REV_NUM_EXT_ID_KEY]))

        # > Update

        new_obj_name = 'routertest_updated'
        with mock.patch.object(self._l3_ovn_client, 'update_router'):
            new_neutron_obj = self._update_router_name(neutron_obj['id'],
                                                       new_obj_name)

        # Assert the revision numbers are out-of-sync
        ovn_obj = self._find_router_row_by_name(obj_name)
        self.assertNotEqual(
            new_neutron_obj['revision_number'],
            int(ovn_obj.external_ids[ovn_const.OVN_REV_NUM_EXT_ID_KEY]))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the old name doesn't exist anymore in the OVNDB
        self.assertIsNone(self._find_router_row_by_name(obj_name))

        # Assert the router is now in sync
        ovn_obj = self._find_router_row_by_name(new_obj_name)
        self.assertEqual(
            new_neutron_obj['revision_number'],
            int(ovn_obj.external_ids[ovn_const.OVN_REV_NUM_EXT_ID_KEY]))

        # > Delete

        with mock.patch.object(self._l3_ovn_client, 'delete_router'):
            self._delete('routers', new_neutron_obj['id'])

        # Assert the router still exists in OVNDB
        self.assertIsNotNone(self._find_router_row_by_name(new_obj_name))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the router is now deleted from OVNDB
        self.assertIsNone(self._find_router_row_by_name(new_obj_name))

        # Assert the revision number no longer exists
        self.assertIsNone(db_rev.get_revision_row(
            self.context,
            new_neutron_obj['id']))

    def test_security_group(self):
        with mock.patch.object(self._ovn_client, 'create_security_group'):
            neutron_obj = self._create_security_group()

        # Assert the sg doesn't exist in OVN
        self.assertIsNone(
            self._find_security_group_row_by_id(neutron_obj['id']))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the sg was now created. We don't save the revision number
        # in the Security Group because OVN doesn't support updating it,
        # all we care about is whether it exists or not.
        self.assertIsNotNone(
            self._find_security_group_row_by_id(neutron_obj['id']))

        # > Delete

        with mock.patch.object(self._ovn_client, 'delete_security_group'):
            self._delete('security-groups', neutron_obj['id'])

        # Assert the sg still exists in OVNDB
        self.assertIsNotNone(
            self._find_security_group_row_by_id(neutron_obj['id']))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the sg is now deleted from OVNDB
        self.assertIsNone(
            self._find_security_group_row_by_id(neutron_obj['id']))

        # Assert the revision number no longer exists
        self.assertIsNone(db_rev.get_revision_row(
            self.context,
            neutron_obj['id']))

    def test_security_group_rule(self):
        neutron_sg = self._create_security_group()
        neutron_net = self._create_network('network1')
        self._create_port('portsgtest', neutron_net['id'],
                          security_groups=[neutron_sg['id']])

        with mock.patch.object(self._ovn_client, 'create_security_group_rule'):
            neutron_obj = self._create_security_group_rule(neutron_sg['id'])

        # Assert the sg rule doesn't exist in OVN
        self.assertIsNone(
            self._find_security_group_rule_row_by_id(neutron_obj['id']))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the sg rule was now created. We don't save the revision number
        # in the Security Group because OVN doesn't support updating it,
        # all we care about is whether it exists or not.
        self.assertIsNotNone(
            self._find_security_group_rule_row_by_id(neutron_obj['id']))

        # > Delete

        # FIXME(lucasagomes): Maintenance thread fixing deleted
        # security group rules is currently broken due to:
        # https://bugs.launchpad.net/networking-ovn/+bug/1756123

    def test_router_port(self):
        neutron_net = self._create_network('networktest', external=True)
        neutron_subnet = self._create_subnet('subnettest', neutron_net['id'])
        neutron_router = self._create_router('routertest')

        with mock.patch.object(self._l3_ovn_client, 'create_router_port'):
            with mock.patch('neutron.db.ovn_revision_numbers_db.'
                            'bump_revision'):
                neutron_obj = self._add_router_interface(neutron_router['id'],
                                                         neutron_subnet['id'])

        # Assert the router port doesn't exist in OVN
        self.assertIsNone(
            self._find_router_port_row_by_port_id(neutron_obj['port_id']))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the router port was now created
        self.assertIsNotNone(
            self._find_router_port_row_by_port_id(neutron_obj['port_id']))

        # > Delete

        with mock.patch.object(self._l3_ovn_client, 'delete_router_port'):
            self._remove_router_interface(neutron_router['id'],
                                          neutron_subnet['id'])

        # Assert the router port still exists in OVNDB
        self.assertIsNotNone(
            self._find_router_port_row_by_port_id(neutron_obj['port_id']))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the router port is now deleted from OVNDB
        self.assertIsNone(
            self._find_router_port_row_by_port_id(neutron_obj['port_id']))

        # Assert the revision number no longer exists
        self.assertIsNone(db_rev.get_revision_row(
            self.context,
            neutron_obj['port_id']))

    def test_check_for_igmp_snooping_enabled(self):
        cfg.CONF.set_override('igmp_snooping_enable', False, group='OVS')
        net = self._create_network('net')
        ls = self.nb_api.db_find('Logical_Switch',
            ('name', '=', utils.ovn_name(net['id']))).execute(
            check_error=True)[0]

        self.assertEqual('false', ls['other_config'][ovn_const.MCAST_SNOOP])
        self.assertEqual(
            'false', ls['other_config'][ovn_const.MCAST_FLOOD_UNREGISTERED])

        # Change the value of the configuration
        cfg.CONF.set_override('igmp_snooping_enable', True, group='OVS')

        # Call the maintenance task and check that the value has been
        # updated in the Logical Switch
        self.assertRaises(periodics.NeverAgain,
                          self.maint.check_for_igmp_snoop_support)

        ls = self.nb_api.db_find('Logical_Switch',
            ('name', '=', utils.ovn_name(net['id']))).execute(
            check_error=True)[0]

        self.assertEqual('true', ls['other_config'][ovn_const.MCAST_SNOOP])
        self.assertEqual(
            'false', ls['other_config'][ovn_const.MCAST_FLOOD_UNREGISTERED])

    def test_check_for_aging_settings(self):
        net = self._create_network('net', provider='datacentre')
        ls = self.nb_api.get_lswitch(utils.ovn_name(net['id']))

        self.assertEqual(
            '0', ls.other_config.get(ovn_const.LS_OPTIONS_FDB_AGE_THRESHOLD))

        # Change the value of the configuration
        cfg.CONF.set_override('fdb_age_threshold', 5, group='ovn')

        # Call the maintenance task and check that the value has been
        # updated in the Logical Switch
        self.assertRaises(periodics.NeverAgain,
                          self.maint.check_fdb_aging_settings)

        ls = self.nb_api.get_lswitch(utils.ovn_name(net['id']))

        self.assertEqual(
            '5', ls.other_config.get(ovn_const.LS_OPTIONS_FDB_AGE_THRESHOLD))

    def test_floating_ip(self):
        ext_net = self._create_network('ext_networktest', external=True)
        ext_subnet = self._create_subnet(
            'ext_subnettest',
            ext_net['id'],
            **{'cidr': '100.0.0.0/24',
               'gateway_ip': '100.0.0.254',
               'allocation_pools': [
                   {'start': '100.0.0.2', 'end': '100.0.0.253'}],
               'enable_dhcp': False})
        net1 = self._create_network('network1test', external=False)
        subnet1 = self._create_subnet('subnet1test', net1['id'])
        external_gateway_info = {
            'enable_snat': True,
            'network_id': ext_net['id'],
            'external_fixed_ips': [
                {'ip_address': '100.0.0.2', 'subnet_id': ext_subnet['id']}]}
        router = self._create_router(
            'routertest', external_gateway_info=external_gateway_info)
        self._add_router_interface(router['id'], subnet1['id'])

        p1 = self._create_port('testp1', net1['id'])
        logical_ip = p1['fixed_ips'][0]['ip_address']
        fip_info = {'floatingip': {
            'tenant_id': self._tenant_id,
            'description': 'test_fip',
            'floating_network_id': ext_net['id'],
            'port_id': p1['id'],
            'fixed_ip_address': logical_ip}}

        # > Create
        with mock.patch.object(self._l3_ovn_client, 'create_floatingip'):
            fip = self.l3_plugin.create_floatingip(self.context, fip_info)

        floating_ip_address = fip['floating_ip_address']
        self.assertEqual(router['id'], fip['router_id'])
        self.assertEqual('testp1', fip['port_details']['name'])
        self.assertIsNotNone(self.nb_api.get_lswitch_port(fip['port_id']))

        # Assert the dnat_and_snat rule doesn't exist in OVN
        self.assertIsNone(
            self._find_nat_rule(router['id'], floating_ip_address, logical_ip))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the rule for the fip is now present
        self.assertIsNotNone(
            self._find_nat_rule(router['id'], floating_ip_address, logical_ip))

        # > Update
        p2 = self._create_port('testp2', net1['id'])
        logical_ip = p2['fixed_ips'][0]['ip_address']
        fip_info = {'floatingip': {
            'port_id': p2['id'],
            'fixed_ip_address': logical_ip}}

        with mock.patch.object(self._l3_ovn_client, 'update_floatingip'):
            self.l3_plugin.update_floatingip(self.context, fip['id'], fip_info)

        # Assert the dnat_and_snat rule in OVN is still using p1's address
        stale_nat_rule = self._find_nat_rule(router['id'], floating_ip_address)
        self.assertEqual(p1['fixed_ips'][0]['ip_address'],
                         stale_nat_rule['logical_ip'])

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the rule for the fip is now updated
        self.assertIsNotNone(
            self._find_nat_rule(router['id'], floating_ip_address, logical_ip))

        # > Delete
        with mock.patch.object(self._l3_ovn_client, 'delete_floatingip'):
            self.l3_plugin.delete_floatingip(self.context, fip['id'])

        self.assertRaises(
            lib_l3_exc.FloatingIPNotFound,
            self.l3_plugin.get_floatingip, self.context, fip['id'])

        # Assert the dnat_and_snat rule in OVN is still present
        self.assertIsNotNone(
            self._find_nat_rule(router['id'], floating_ip_address, logical_ip))

        # Call the maintenance thread to fix the problem
        self.maint.check_for_inconsistencies()

        # Assert the rule for the fip is now gone
        self.assertIsNone(
            self._find_nat_rule(router['id'], floating_ip_address))

        # Assert the router snat rule is still there
        snat_rule = self._find_nat_rule(
            router['id'], '100.0.0.2', nat_type='snat')
        self.assertEqual(subnet1['cidr'], snat_rule['logical_ip'])

    def test_port_forwarding(self):
        fip_attrs = lambda args: {
            pf_def.RESOURCE_NAME: {pf_def.RESOURCE_NAME: args}}

        def _verify_lb(test, protocol, vip_ext_port, vip_int_port):
            ovn_lbs = self._find_pf_lb(router_id, fip_id)
            test.assertEqual(1, len(ovn_lbs))
            test.assertEqual('pf-floatingip-{}-{}'.format(fip_id, protocol),
                             ovn_lbs[0].name)
            test.assertEqual(
                {'{}:{}'.format(fip_ip, vip_ext_port):
                 '{}:{}'.format(p1_ip, vip_int_port)},
                ovn_lbs[0].vips)

        ext_net = self._create_network('ext_networktest', external=True)
        ext_subnet = self._create_subnet(
            'ext_subnettest',
            ext_net['id'],
            **{'cidr': '100.0.0.0/24',
               'gateway_ip': '100.0.0.254',
               'allocation_pools': [
                   {'start': '100.0.0.2', 'end': '100.0.0.253'}],
               'enable_dhcp': False})
        net1 = self._create_network('network1test', external=False)
        subnet1 = self._create_subnet('subnet1test', net1['id'])
        external_gateway_info = {
            'enable_snat': True,
            'network_id': ext_net['id'],
            'external_fixed_ips': [
                {'ip_address': '100.0.0.2', 'subnet_id': ext_subnet['id']}]}
        router = self._create_router(
            'routertest', external_gateway_info=external_gateway_info)
        router_id = router['id']
        self._add_router_interface(router['id'], subnet1['id'])

        fip_info = {'floatingip': {
            'tenant_id': self._tenant_id,
            'floating_network_id': ext_net['id'],
            'port_id': None,
            'fixed_ip_address': None}}
        fip = self.l3_plugin.create_floatingip(self.context, fip_info)
        fip_id = fip['id']
        fip_ip = fip['floating_ip_address']
        p1 = self._create_port('testp1', net1['id'])
        p1_ip = p1['fixed_ips'][0]['ip_address']

        with mock.patch('neutron_lib.callbacks.registry.publish') as m_publish:
            # > Create
            fip_pf_args = {
                pf_def.EXTERNAL_PORT: 2222,
                pf_def.INTERNAL_PORT: 22,
                pf_def.INTERNAL_PORT_ID: p1['id'],
                pf_def.PROTOCOL: 'tcp',
                pf_def.INTERNAL_IP_ADDRESS: p1_ip}
            pf_obj = self.pf_plugin.create_floatingip_port_forwarding(
                self.context, fip_id, **fip_attrs(fip_pf_args))
            call = mock.call('port_forwarding', 'after_create', self.pf_plugin,
                             payload=mock.ANY)
            m_publish.assert_has_calls([call])

            # Assert load balancer for port forwarding was not created
            self.assertFalse(self._find_pf_lb(router_id, fip_id))

            # Call the maintenance thread to fix the problem
            self.maint.check_for_inconsistencies()

            # Assert load balancer for port forwarding was created
            _verify_lb(self, 'tcp', 2222, 22)

            # > Update
            fip_pf_args = {pf_def.EXTERNAL_PORT: 5353,
                           pf_def.INTERNAL_PORT: 53,
                           pf_def.PROTOCOL: 'udp'}
            m_publish.reset_mock()
            self.pf_plugin.update_floatingip_port_forwarding(
                self.context, pf_obj['id'], fip_id, **fip_attrs(fip_pf_args))
            call = mock.call('port_forwarding', 'after_update', self.pf_plugin,
                             payload=mock.ANY)
            m_publish.assert_has_calls([call])

            # Assert load balancer for port forwarding is stale
            _verify_lb(self, 'tcp', 2222, 22)

            # Call the maintenance thread to fix the problem
            self.maint.check_for_inconsistencies()

            # Assert load balancer for port forwarding was updated
            _verify_lb(self, 'udp', 5353, 53)

            # > Delete
            m_publish.reset_mock()
            self.pf_plugin.delete_floatingip_port_forwarding(
                self.context, pf_obj['id'], fip_id)
            call = mock.call('port_forwarding', 'after_delete', self.pf_plugin,
                             payload=mock.ANY)
            m_publish.assert_has_calls([call])

            # Assert load balancer for port forwarding is stale
            _verify_lb(self, 'udp', 5353, 53)

            # Call the maintenance thread to fix the problem
            self.maint.check_for_inconsistencies()

            # Assert load balancer for port forwarding is gone
            self.assertFalse(self._find_pf_lb(router_id, fip_id))

    def test_check_for_ha_chassis_group(self):
        net1 = self._create_network('network1test', external=False)
        self._create_subnet('subnet1test', net1['id'])
        p1 = self._create_port('testp1', net1['id'], vnic_type='direct')

        # Remove the HA Chassis Group register, created during the port
        # creation.
        self.nb_api.set_lswitch_port(p1['id'], ha_chassis_group=[]).execute(
            check_error=True)
        hcg_uuid = next(iter(self.nb_api._tables['HA_Chassis_Group'].rows))
        self.nb_api.ha_chassis_group_del(hcg_uuid).execute(check_error=True)
        lsp = self.nb_api.lookup('Logical_Switch_Port', p1['id'])
        self.assertEqual([], lsp.ha_chassis_group)

        self.assertRaises(periodics.NeverAgain,
                          self.maint.check_for_ha_chassis_group)
        hcg_uuid = next(iter(self.nb_api._tables['HA_Chassis_Group'].rows))
        lsp = self.nb_api.lookup('Logical_Switch_Port', p1['id'])
        self.assertEqual(hcg_uuid, lsp.ha_chassis_group[0].uuid)

    def test_remove_duplicated_chassis_registers(self):
        hostnames = ['host1', 'host2']
        for hostname in hostnames:
            for _ in range(3):
                self.add_fake_chassis(hostname)

        chassis = self.sb_api.chassis_list().execute(check_error=True)
        self.assertEqual(6, len(chassis))
        # Make the chassis private timestamp different
        for idx, ch in enumerate(chassis):
            self.sb_api.db_set('Chassis_Private', ch.name,
                               ('nb_cfg_timestamp', idx)).execute()

        ch_private_dict = {}  # host: [ch_private1, ch_private2, ...]
        for hostname in hostnames:
            ch_private_list = []
            for ch in (ch for ch in chassis if ch.hostname == hostname):
                ch_private = self.sb_api.lookup('Chassis_Private', ch.name,
                                                default=None)
                if ch_private:
                    # One of the "Chassis_Private" has been deleted on purpose
                    # in this test.
                    ch_private_list.append(ch_private)
            ch_private_list.sort(key=lambda x: x.nb_cfg_timestamp,
                                 reverse=True)
            ch_private_dict[hostname] = ch_private_list

        self.maint.remove_duplicated_chassis_registers()
        chassis_result = self.sb_api.chassis_list().execute(check_error=True)
        self.assertEqual(2, len(chassis_result))
        for ch in chassis_result:
            self.assertIn(ch.hostname, hostnames)
            hostnames.remove(ch.hostname)
            # From ch_private_dict[ch.hostname], we retrieve the first
            # "Chassis_Private" register because these are ordered by
            # timestamp. The newer one (bigger timestamp) should remain in the
            # system.
            ch_expected = ch_private_dict[ch.hostname][0].chassis[0]
            self.assertEqual(ch_expected.name, ch.name)

    def test_remove_duplicated_chassis_registers_no_ch_private_register(self):
        for _ in range(2):
            self.add_fake_chassis('host1')

        chassis = self.sb_api.chassis_list().execute(check_error=True)
        self.assertEqual(2, len(chassis))
        # Make the chassis private timestamp different
        # Delete on of the "Chassis_Private" registers.
        self.sb_api.db_destroy('Chassis_Private', chassis[0].name).execute()
        self.sb_api.db_set('Chassis_Private', chassis[1].name,
                           ('nb_cfg_timestamp', 1)).execute()

        self.maint.remove_duplicated_chassis_registers()
        chassis_result = self.sb_api.chassis_list().execute(check_error=True)
        # Both "Chassis" registers are still in the DB because one
        # "Chassis_Private" register was missing.
        self.assertEqual(2, len(chassis_result))

    def test_configure_nb_global(self):
        def options_intersect(options1, options2):
            return bool(set(
                new_nb_global_options.keys()).intersection(nb_options.keys()))

        new_nb_global_options = {
            'foo': 'bar',
            'baz': 'qux',
        }

        cfg_nb_global_options = [
            ovn_config.cfg.StrOpt(key) for key in new_nb_global_options
        ]

        def get_opt(key):
            return new_nb_global_options[key]

        nb_options = self.nb_api.db_get('NB_Global', '.', 'options').execute(
            check_error=True, log_errors=True)
        self.assertFalse(options_intersect(new_nb_global_options, nb_options))

        with mock.patch.object(
                ovn_config, 'nb_global_opts', cfg_nb_global_options), \
                mock.patch.object(
                    ovn_config.cfg.CONF.ovn_nb_global, 'get',
                    side_effect=get_opt):
            self.assertRaises(periodics.NeverAgain,
                              self.maint.configure_nb_global)

        nb_options = self.nb_api.db_get('NB_Global', '.', 'options').execute(
            check_error=True, log_errors=True)
        self.assertTrue(options_intersect(new_nb_global_options, nb_options))

    def test_floating_ip_with_gateway_port(self):
        ext_net = self._create_network('ext_networktest', external=True)
        ext_subnet = self._create_subnet(
            'ext_subnettest',
            ext_net['id'],
            **{'cidr': '100.0.0.0/24',
               'gateway_ip': '100.0.0.254',
               'allocation_pools': [
                   {'start': '100.0.0.2', 'end': '100.0.0.253'}],
               'enable_dhcp': False})
        net1 = self._create_network('network1test', external=False)
        subnet1 = self._create_subnet('subnet1test', net1['id'])
        external_gateway_info = {
            'enable_snat': True,
            'network_id': ext_net['id'],
            'external_fixed_ips': [
                {'ip_address': '100.0.0.2', 'subnet_id': ext_subnet['id']}]}
        router = self._create_router(
            'routertest', external_gateway_info=external_gateway_info)
        self._add_router_interface(router['id'], subnet1['id'])

        p1 = self._create_port('testp1', net1['id'])
        logical_ip = p1['fixed_ips'][0]['ip_address']
        fip_info = {'floatingip': {
            'tenant_id': self._tenant_id,
            'description': 'test_fip',
            'floating_network_id': ext_net['id'],
            'port_id': p1['id'],
            'fixed_ip_address': logical_ip}}

        # Create floating IP without gateway_port
        with mock.patch.object(utils,
                'is_nat_gateway_port_supported', return_value=False):
            fip = self.l3_plugin.create_floatingip(self.context, fip_info)

        self.assertEqual(router['id'], fip['router_id'])
        self.assertEqual('testp1', fip['port_details']['name'])
        self.assertIsNotNone(self.nb_api.get_lswitch_port(fip['port_id']))

        rules = self.nb_api.get_all_logical_routers_with_rports()[0]
        fip_rule = rules['dnat_and_snats'][0]
        if utils.is_nat_gateway_port_supported(self.nb_api):
            self.assertEqual([], fip_rule['gateway_port'])
        else:
            self.assertNotIn('gateway_port', fip_rule)

        # Call the maintenance task and check that the value has been
        # updated in the NAT rule
        self.assertRaises(periodics.NeverAgain,
            self.maint.update_nat_floating_ip_with_gateway_port_reference)

        rules = self.nb_api.get_all_logical_routers_with_rports()[0]
        fip_rule = rules['dnat_and_snats'][0]

        if utils.is_nat_gateway_port_supported(self.nb_api):
            self.assertNotEqual([], fip_rule['gateway_port'])
        else:
            self.assertNotIn('gateway_port', fip_rule)


class TestLogMaintenance(_TestMaintenanceHelper,
                         test_log_driver.LogApiTestCaseBase):
    def test_check_for_logging_conf_change(self):
        # Check logging is supported
        if not self.log_driver.network_logging_supported(self.nb_api):
            self.skipTest("The current OVN version does not offer support "
                          "for neutron network log functionality.")
            self.assertIsNotNone(self.log_plugin)
        # Check no meter exists
        self.assertFalse(self.nb_api._tables['Meter'].rows.values())
        # Add a log object
        self.log_plugin.create_log(self.context, self._log_data())
        # Check a meter and fair meter exist
        self.assertTrue(self.nb_api._tables['Meter'].rows)
        self.assertTrue(self.nb_api._tables['Meter_Band'].rows)
        self.assertEqual(len([*self.nb_api._tables['Meter'].rows.values()]),
            len([*self.nb_api._tables['Meter_Band'].rows.values()]))
        self._check_meters_consistency()
        # Update burst and rate limit values on the configuration
        ovn_config.cfg.CONF.set_override('burst_limit', CFG_NEW_BURST,
                                         group='network_log')
        ovn_config.cfg.CONF.set_override('rate_limit', CFG_NEW_RATE,
                                         group='network_log')
        # Call the maintenance task
        self.assertRaises(periodics.NeverAgain,
                          self.maint.check_fair_meter_consistency)
        # Check meter band was effectively changed after the maintenance call
        self._check_meters_consistency(CFG_NEW_BURST, CFG_NEW_RATE)

    def _check_meters_consistency(self, new_burst=None, new_rate=None):
        burst, rate = (new_burst, new_rate) if new_burst else (
            cfg.CONF.network_log.burst_limit, cfg.CONF.network_log.rate_limit)
        for meter in [*self.nb_api._tables['Meter'].rows.values()]:
            meter_band = self.nb_api.lookup('Meter_Band', meter.bands[0].uuid)
            if "_stateless" in meter.name:
                self.assertEqual(int(burst / 2), meter_band.burst_size)
                self.assertEqual(int(rate / 2), meter_band.rate)
            else:
                self.assertEqual(burst, meter_band.burst_size)
                self.assertEqual(rate, meter_band.rate)



# Copyright (c) 2016 Intel Corporation.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

import netaddr

from neutron.common import utils as n_utils
from neutron.objects import base


class EndpointBase(base.NeutronDbObject):

    primary_keys = ['ip_address']

    @classmethod
    def modify_fields_from_db(cls, db_obj):
        result = super(EndpointBase, cls).modify_fields_from_db(db_obj)
        if 'ip_address' in result:
            result['ip_address'] = netaddr.IPAddress(result['ip_address'])
        return result

    @classmethod
    def modify_fields_to_db(cls, fields):
        result = super(EndpointBase, cls).modify_fields_to_db(fields)
        if 'ip_address' in fields:
            result['ip_address'] = cls.filter_to_str(result['ip_address'])
        return result


class SegmentAllocation(object, metaclass=abc.ABCMeta):

    @classmethod
    def get_random_unallocated_segment(cls, context, **filters):
        with cls.db_context_reader(context):
            columns = set(dict(cls.db_model.__table__.columns))
            model_filters = dict((k, filters[k])
                                 for k in columns & set(filters.keys()))
            query = context.session.query(cls.db_model).filter_by(
                allocated=False, **model_filters)
            rand_func = n_utils.get_sql_random_method(
                context.session.bind.dialect.name)
            if rand_func:
                query = query.order_by(rand_func())
            return query.first()

    @classmethod
    def allocate(cls, context, **segment):
        with cls.db_context_writer(context):
            return context.session.query(cls.db_model).filter_by(
                allocated=False, **segment).update({'allocated': True})

    @classmethod
    def deallocate(cls, context, **segment):
        with cls.db_context_writer(context):
            return context.session.query(cls.db_model).filter_by(
                allocated=True, **segment).update({'allocated': False})

    @classmethod
    def update_primary_keys(cls, _dict, segmentation_id=None, **kwargs):
        _dict[cls.primary_keys[0]] = segmentation_id

    @abc.abstractmethod
    def get_segmentation_id(self):
        pass

    @property
    def segmentation_id(self):
        return self.db_obj.segmentation_id



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from neutron_lib import policy as neutron_policy
from oslo_log import versionutils
from oslo_policy import policy

from neutron.conf.policies import base


DEPRECATED_REASON = """
The floating IP port forwarding API now supports system scope and default
roles.
"""

COLLECTION_PATH = '/floatingips/{floatingip_id}/port_forwardings'
RESOURCE_PATH = ('/floatingips/{floatingip_id}'
                 '/port_forwardings/{port_forwarding_id}')


rules = [
    policy.DocumentedRuleDefault(
        name='create_floatingip_port_forwarding',
        check_str=base.ADMIN_OR_PARENT_OWNER_MEMBER,
        scope_types=['project'],
        description='Create a floating IP port forwarding',
        operations=[
            {
                'method': 'POST',
                'path': COLLECTION_PATH,
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='create_floatingip_port_forwarding',
            check_str=neutron_policy.RULE_ADMIN_OR_PARENT_OWNER,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='get_floatingip_port_forwarding',
        check_str=base.ADMIN_OR_PARENT_OWNER_READER,
        scope_types=['project'],
        description='Get a floating IP port forwarding',
        operations=[
            {
                'method': 'GET',
                'path': COLLECTION_PATH,
            },
            {
                'method': 'GET',
                'path': RESOURCE_PATH,
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='get_floatingip_port_forwarding',
            check_str=neutron_policy.RULE_ADMIN_OR_PARENT_OWNER,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='update_floatingip_port_forwarding',
        check_str=base.ADMIN_OR_PARENT_OWNER_MEMBER,
        scope_types=['project'],
        description='Update a floating IP port forwarding',
        operations=[
            {
                'method': 'PUT',
                'path': RESOURCE_PATH,
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='update_floatingip_port_forwarding',
            check_str=neutron_policy.RULE_ADMIN_OR_PARENT_OWNER,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='delete_floatingip_port_forwarding',
        check_str=base.ADMIN_OR_PARENT_OWNER_MEMBER,
        scope_types=['project'],
        description='Delete a floating IP port forwarding',
        operations=[
            {
                'method': 'DELETE',
                'path': RESOURCE_PATH,
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='delete_floatingip_port_forwarding',
            check_str=neutron_policy.RULE_ADMIN_OR_PARENT_OWNER,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
]


def list_rules():
    return rules



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from unittest import mock

from neutron_lib import constants
from neutron_lib import context
from neutron_lib.services.qos import constants as qos_consts
from oslo_utils import uuidutils

from neutron.objects.qos import policy
from neutron.objects.qos import rule
from neutron.plugins.ml2.common import constants as comm_consts
from neutron.plugins.ml2.drivers.openvswitch.agent import (
        ovs_agent_extension_api as ovs_ext_api)
from neutron.plugins.ml2.drivers.openvswitch.agent.extension_drivers import (
    qos_driver)
from neutron.plugins.ml2.drivers.openvswitch.agent.openflow.native import (
    ovs_bridge)
from neutron.tests.unit.plugins.ml2.drivers.openvswitch.agent import (
    ovs_test_base)


class QosOVSAgentDriverTestCase(ovs_test_base.OVSAgentConfigTestBase):

    def setUp(self):
        super(QosOVSAgentDriverTestCase, self).setUp()
        conn_patcher = mock.patch(
            'neutron.agent.ovsdb.impl_idl._connection')
        conn_patcher.start()
        self.addCleanup(conn_patcher.stop)
        self.context = context.get_admin_context()
        self.qos_driver = qos_driver.QosOVSAgentDriver()
        self.mock_clear_minimum_bandwidth_qos = mock.patch.object(
            self.qos_driver, '_qos_bandwidth_initialize').start()
        os_ken_app = mock.Mock()
        self.agent_api = ovs_ext_api.OVSAgentExtensionAPI(
                         ovs_bridge.OVSAgentBridge(
                             'br-int', os_ken_app=os_ken_app),
                         ovs_bridge.OVSAgentBridge(
                             'br-tun', os_ken_app=os_ken_app),
                         {'phys1': ovs_bridge.OVSAgentBridge(
                             'br-phys1', os_ken_app=os_ken_app)})
        self.qos_driver.consume_api(self.agent_api)
        mock.patch.object(
            qos_driver.MeterRuleManager, '_init_max_meter_id').start()
        self.qos_driver.initialize()
        self.qos_driver.br_int = mock.Mock()
        self.qos_driver.br_int.get_dp = mock.Mock(return_value=(mock.Mock(),
                                                                mock.Mock(),
                                                                mock.Mock()))
        self.qos_driver.meter_cache_pps.br_int = self.qos_driver.br_int
        self.qos_driver.meter_cache_pps.max_meter = 65535
        self.qos_driver.br_int.list_meter_features = mock.Mock(
            return_value=[{"max_meter": 65535,
                           "band_types": 2,
                           "capabilities": 15,
                           "max_bands": 8}])
        self.qos_driver.br_int.get_egress_bw_limit_for_port = mock.Mock(
            return_value=(1000, 10))
        self.get_egress = self.qos_driver.br_int.get_egress_bw_limit_for_port
        self.qos_driver.br_int.del_egress_bw_limit_for_port = mock.Mock()
        self.delete_egress = (
            self.qos_driver.br_int.delete_egress_bw_limit_for_port)
        self.delete_ingress = (
            self.qos_driver.br_int.delete_ingress_bw_limit_for_port)
        self.create_egress = (
            self.qos_driver.br_int.create_egress_bw_limit_for_port)
        self.update_ingress = (
            self.qos_driver.br_int.update_ingress_bw_limit_for_port)

        self.apply_meter_to_port = (
            self.qos_driver.br_int.apply_meter_to_port)
        self.remove_meter_from_port = (
            self.qos_driver.br_int.remove_meter_from_port)
        self.delete_meter = (
            self.qos_driver.br_int.delete_meter)
        self.create_meter = (
            self.qos_driver.br_int.create_meter)
        self.update_meter = (
            self.qos_driver.br_int.update_meter)

        self.rules = [
            self._create_bw_limit_rule_obj(constants.EGRESS_DIRECTION),
            self._create_bw_limit_rule_obj(constants.INGRESS_DIRECTION),
            self._create_pps_limit_rule_obj(constants.EGRESS_DIRECTION),
            self._create_pps_limit_rule_obj(constants.INGRESS_DIRECTION),
            self._create_dscp_marking_rule_obj()]
        self.qos_policy = self._create_qos_policy_obj(self.rules)
        self.port = self._create_fake_port(self.qos_policy.id)
        self.qos_driver.br_int.get_port_tag_by_name = mock.Mock(
            return_value=1)
        self.addCleanup(self._reset_meter_id_generator_singleton)

    @staticmethod
    def _reset_meter_id_generator_singleton():
        if hasattr(qos_driver.MeterIDGenerator, '_instance'):
            del(qos_driver.MeterIDGenerator._instance)

    def _create_bw_limit_rule_obj(self, direction):
        rule_obj = rule.QosBandwidthLimitRule()
        rule_obj.id = uuidutils.generate_uuid()
        rule_obj.max_kbps = 2
        rule_obj.max_burst_kbps = 200
        rule_obj.direction = direction
        rule_obj.obj_reset_changes()
        return rule_obj

    def _create_dscp_marking_rule_obj(self):
        rule_obj = rule.QosDscpMarkingRule()
        rule_obj.id = uuidutils.generate_uuid()
        rule_obj.dscp_mark = 32
        rule_obj.obj_reset_changes()
        return rule_obj

    def _create_pps_limit_rule_obj(self, direction):
        rule_obj = rule.QosPacketRateLimitRule()
        rule_obj.id = uuidutils.generate_uuid()
        rule_obj.max_kpps = 2000
        rule_obj.max_burst_kpps = 200
        rule_obj.direction = direction
        rule_obj.obj_reset_changes()
        return rule_obj

    def _create_qos_policy_obj(self, rules):
        policy_dict = {'id': uuidutils.generate_uuid(),
                'project_id': uuidutils.generate_uuid(),
                'name': 'test',
                'description': 'test',
                'shared': False,
                'rules': rules}
        policy_obj = policy.QosPolicy(self.context, **policy_dict)
        policy_obj.obj_reset_changes()
        for policy_rule in policy_obj.rules:
            policy_rule.qos_policy_id = policy_obj.id
            policy_rule.obj_reset_changes()
        return policy_obj

    def _create_fake_port(self, policy_id):
        self.port_name = 'fakeport'

        port_id = uuidutils.generate_uuid()

        class FakeVifPort(object):
            port_name = self.port_name
            ofport = 111
            vif_id = port_id
            vif_mac = "aa:bb:cc:dd:ee:ff"

        return {'vif_port': FakeVifPort(),
                'qos_policy_id': policy_id,
                'qos_network_policy_id': None,
                'port_id': port_id,
                'device_owner': uuidutils.generate_uuid()}

    def test_create_new_rules(self):
        self.qos_driver.br_int.get_value_from_other_config = mock.Mock()
        self.qos_driver.br_int.set_value_to_other_config = mock.Mock()

        self.qos_driver.br_int.get_egress_bw_limit_for_port = mock.Mock(
            return_value=(None, None))
        self.qos_driver.create(self.port, self.qos_policy)
        self.assertEqual(0, self.delete_egress.call_count)
        self.assertEqual(0, self.delete_ingress.call_count)
        self.create_egress.assert_called_once_with(
            self.port_name, self.rules[0].max_kbps,
            self.rules[0].max_burst_kbps)
        self.update_ingress.assert_called_once_with(
            self.port_name, self.rules[1].max_kbps,
            self.rules[1].max_burst_kbps)
        self._assert_dscp_rule_create_updated()

        self.create_meter.assert_has_calls(
            [mock.call(mock.ANY, self.rules[2].max_kpps * 1000,
             burst=self.rules[2].max_burst_kpps * 1000,
             type_=comm_consts.METER_FLAG_PPS),
             mock.call(mock.ANY, self.rules[3].max_kpps * 1000,
             burst=self.rules[3].max_burst_kpps * 1000,
             type_=comm_consts.METER_FLAG_PPS)])
        self.apply_meter_to_port.assert_has_calls(
            [mock.call(mock.ANY, constants.EGRESS_DIRECTION,
                       "aa:bb:cc:dd:ee:ff",
                       in_port=111,
                       type_=comm_consts.METER_FLAG_PPS),
             mock.call(mock.ANY, constants.INGRESS_DIRECTION,
                       "aa:bb:cc:dd:ee:ff",
                       local_vlan=1,
                       type_=comm_consts.METER_FLAG_PPS)])

    def test_create_existing_rules(self):
        self.qos_driver.create(self.port, self.qos_policy)
        self._assert_rules_create_updated()
        self._assert_dscp_rule_create_updated()

    def test_update_rules(self):
        self.qos_driver.update(self.port, self.qos_policy)
        self._assert_rules_create_updated()
        self._assert_dscp_rule_create_updated()

    def test_update_rules_no_vif_port(self):
        port = copy.copy(self.port)
        port.pop("vif_port")
        self.qos_driver.update(port, self.qos_policy)
        self.create_egress.assert_not_called()
        self.update_ingress.assert_not_called()

        self.create_meter.assert_not_called()
        self.apply_meter_to_port.assert_not_called()

    def _test_delete_rules(self, qos_policy):
        self.qos_driver.create(self.port, qos_policy)
        self.qos_driver.delete(self.port, qos_policy)
        self.delete_egress.assert_called_once_with(self.port_name)
        self.delete_ingress.assert_called_once_with(self.port_name)

        self.assertEqual(2, self.delete_meter.call_count)
        self.remove_meter_from_port.assert_has_calls(
            [mock.call(constants.EGRESS_DIRECTION,
                       "aa:bb:cc:dd:ee:ff",
                       in_port=111,
                       type_=comm_consts.METER_FLAG_PPS),
             mock.call(constants.INGRESS_DIRECTION,
                       "aa:bb:cc:dd:ee:ff",
                       local_vlan=1,
                       type_=comm_consts.METER_FLAG_PPS)])

    def _test_delete_rules_no_policy(self):
        self.qos_driver.delete(self.port)
        self.delete_egress.assert_called_once_with(self.port_name)
        self.delete_ingress.assert_called_once_with(self.port_name)

    def test_delete_rules(self):
        self._test_delete_rules(self.qos_policy)

    def test_delete_rules_no_policy(self):
        self._test_delete_rules_no_policy()

    def test_delete_rules_no_vif_port(self):
        port = copy.copy(self.port)
        port.pop("vif_port")
        self.qos_driver.delete(port, self.qos_policy)
        self.delete_egress.assert_not_called()
        self.delete_ingress.assert_not_called()
        self.delete_meter.assert_not_called()

        self.delete_meter.assert_not_called()
        self.remove_meter_from_port.assert_not_called()

    def test_meter_manager_allocate_meter_id(self):
        meter_cache_pps = qos_driver.MeterRuleManager(mock.Mock())
        meter_cache_pps.generator.max_meter = 10000
        meter_cache_bps = qos_driver.MeterRuleManager(
            mock.Mock(), type_=comm_consts.METER_FLAG_BPS)
        meter_cache_bps.generator.max_meter = 10000
        meter_cache_pps.allocate_meter_id("1", "ingress")
        meter_cache_pps.allocate_meter_id("1", "egress")
        meter_cache_bps.allocate_meter_id("1", "ingress")
        meter_cache_bps.allocate_meter_id("1", "egress")
        meter_cache_pps.allocate_meter_id("2", "ingress")
        meter_cache_pps.allocate_meter_id("2", "egress")
        meter_cache_bps.allocate_meter_id("2", "ingress")
        meter_cache_bps.allocate_meter_id("2", "egress")
        self.assertEqual(
            meter_cache_pps.generator.PORT_METER_ID,
            meter_cache_bps.generator.PORT_METER_ID)
        pps_keys = meter_cache_pps.generator.PORT_METER_ID.keys()
        bps_keys = meter_cache_bps.generator.PORT_METER_ID.keys()
        self.assertEqual(
            2, len([k for k in pps_keys if k.startswith('pps_1')]))
        self.assertEqual(
            2, len([k for k in bps_keys if k.startswith('bps_1')]))
        self.assertEqual(
            2, len([k for k in pps_keys if k.startswith('pps_2')]))
        self.assertEqual(
            2, len([k for k in bps_keys if k.startswith('bps_2')]))
        self.assertEqual(
            meter_cache_pps.generator.PORT_METER_ID.keys(),
            meter_cache_bps.generator.PORT_METER_ID.keys())
        pps_values = list(meter_cache_pps.generator.PORT_METER_ID.values())
        bps_values = list(meter_cache_bps.generator.PORT_METER_ID.values())
        self.assertEqual(pps_values, bps_values)
        except_keys = ["pps_1_ingress", "pps_1_egress",
                       "bps_1_ingress", "bps_1_egress",
                       "pps_2_ingress", "pps_2_egress",
                       "bps_2_ingress", "bps_2_egress"]
        except_values = []
        for key in except_keys:
            value = meter_cache_bps.generator.PORT_METER_ID.get(key)
            if value:
                except_values.append(value)
        self.assertEqual(8, len(set(except_values)))

    def test_meter_manager_remove_port_meter_id(self):
        meter_cache_pps = qos_driver.MeterRuleManager(mock.Mock())
        meter_cache_pps.generator.max_meter = 10000
        meter_cache_bps = qos_driver.MeterRuleManager(
            mock.Mock(), type_=comm_consts.METER_FLAG_BPS)
        meter_cache_bps.generator.max_meter = 10000
        meter_cache_pps.allocate_meter_id("1", "ingress")
        meter_cache_pps.allocate_meter_id("1", "egress")
        meter_cache_bps.allocate_meter_id("1", "ingress")
        meter_cache_bps.allocate_meter_id("1", "egress")
        meter_cache_pps.allocate_meter_id("2", "ingress")
        meter_cache_pps.allocate_meter_id("2", "egress")
        meter_cache_bps.allocate_meter_id("2", "ingress")
        meter_cache_bps.allocate_meter_id("2", "egress")
        self.assertEqual(
            meter_cache_pps.generator.PORT_METER_ID,
            meter_cache_bps.generator.PORT_METER_ID)

        meter_cache_bps.remove_port_meter_id("2", "ingress")
        meter_cache_pps.remove_port_meter_id("1", "egress")

        self.assertNotIn(
            "pps_1_egress", meter_cache_pps.generator.PORT_METER_ID.keys())
        self.assertNotIn(
            "bps_2_ingress", meter_cache_pps.generator.PORT_METER_ID.keys())

        pps_values = list(meter_cache_pps.generator.PORT_METER_ID.values())
        bps_values = list(meter_cache_bps.generator.PORT_METER_ID.values())
        self.assertEqual(pps_values, bps_values)
        except_keys = ["pps_1_ingress",
                       "bps_1_ingress", "bps_1_egress",
                       "pps_2_ingress", "pps_2_egress",
                       "bps_2_egress"]
        except_values = []
        for key in except_keys:
            value = meter_cache_bps.generator.PORT_METER_ID.get(key)
            if value:
                except_values.append(value)
        self.assertEqual(6, len(set(except_values)))

    def _assert_rules_create_updated(self):
        self.create_egress.assert_called_once_with(
            self.port_name, self.rules[0].max_kbps,
            self.rules[0].max_burst_kbps)
        self.update_ingress.assert_called_once_with(
            self.port_name, self.rules[1].max_kbps,
            self.rules[1].max_burst_kbps)

        self.create_meter.assert_has_calls(
            [mock.call(mock.ANY, self.rules[2].max_kpps * 1000,
             burst=self.rules[2].max_burst_kpps * 1000,
             type_=comm_consts.METER_FLAG_PPS),
             mock.call(mock.ANY, self.rules[3].max_kpps * 1000,
             burst=self.rules[3].max_burst_kpps * 1000,
             type_=comm_consts.METER_FLAG_PPS)])
        self.apply_meter_to_port.assert_has_calls(
            [mock.call(mock.ANY, constants.EGRESS_DIRECTION,
                       "aa:bb:cc:dd:ee:ff",
                       in_port=111,
                       type_=comm_consts.METER_FLAG_PPS),
             mock.call(mock.ANY, constants.INGRESS_DIRECTION,
                       "aa:bb:cc:dd:ee:ff",
                       local_vlan=1,
                       type_=comm_consts.METER_FLAG_PPS)])

    def _assert_dscp_rule_create_updated(self):
        # Assert install_instructions is the last call
        self.assertEqual(
            'install_dscp_marking_rule',
            self.qos_driver.br_int.method_calls[-1][0])

        self.qos_driver.br_int.install_dscp_marking_rule.\
            assert_called_once_with(dscp_mark=mock.ANY, port=mock.ANY)

    def test_create_minimum_bandwidth(self):
        with mock.patch.object(self.qos_driver, 'update_minimum_bandwidth') \
                as mock_update_minimum_bandwidth:
            self.qos_driver.create_minimum_bandwidth('port_name', 'rule')
            mock_update_minimum_bandwidth.assert_called_once_with('port_name',
                                                                  'rule')

    def test_delete_minimum_bandwidth(self):
        with mock.patch.object(self.qos_driver.br_int,
                               'delete_minimum_bandwidth_queue') \
                as mock_delete_minimum_bandwidth_queue:
            self.qos_driver.ports['p_id'] = {}
            self.qos_driver.delete_minimum_bandwidth({'port_id': 'p_id'})
            mock_delete_minimum_bandwidth_queue.assert_not_called()

            mock_delete_minimum_bandwidth_queue.reset_mock()
            self.qos_driver.ports['p_id'] = {
                (qos_consts.RULE_TYPE_MINIMUM_BANDWIDTH,
                 constants.EGRESS_DIRECTION): 'rule_port'}
            self.qos_driver.delete_minimum_bandwidth({'port_id': 'p_id'})
            mock_delete_minimum_bandwidth_queue.assert_called_once_with('p_id')

    @mock.patch.object(qos_driver, 'LOG')
    def test_update_minimum_bandwidth_no_vif_port(self, mock_log):
        with mock.patch.object(self.qos_driver.br_int,
                               'update_minimum_bandwidth_queue') \
                as mock_delete_minimum_bandwidth_queue:
            self.qos_driver.update_minimum_bandwidth(
                {'port_id': 'portid'}, mock.ANY)
            mock_delete_minimum_bandwidth_queue.assert_not_called()
            mock_log.debug.assert_called_once_with(
                'update_minimum_bandwidth was received for port %s but '
                'vif_port was not found. It seems that port is already '
                'deleted', 'portid')

    @mock.patch.object(qos_driver, 'LOG')
    def test_update_minimum_bandwidth_no_physical_network(self, mock_log):
        with mock.patch.object(self.qos_driver.br_int,
                               'update_minimum_bandwidth_queue') \
                as mock_delete_minimum_bandwidth_queue:
            port = {'vif_port': mock.ANY, 'port_id': 'portid',
                    'physical_network': None}
            self.qos_driver.update_minimum_bandwidth(port, mock.ANY)
            mock_delete_minimum_bandwidth_queue.assert_not_called()
            mock_log.debug.assert_called_once_with(
                'update_minimum_bandwidth was received for port %s but '
                'has no physical network associated', 'portid')

    def test_update_minimum_bandwidth_no_phy_brs(self):
        vif_port = mock.Mock()
        vif_port.ofport = 'ofport'
        rule = mock.Mock()
        rule.min_kbps = 1500
        port = {'port_id': 'port_id', 'vif_port': vif_port,
                'physical_network': mock.ANY}
        with mock.patch.object(self.qos_driver.br_int,
                               'update_minimum_bandwidth_queue') \
                as mock_delete_minimum_bandwidth_queue, \
                mock.patch.object(self.qos_driver.agent_api,
                                  'request_phy_brs'):
            self.qos_driver.update_minimum_bandwidth(port, rule)
            mock_delete_minimum_bandwidth_queue.assert_called_once_with(
                'port_id', [], 'ofport', 1500)

    def test_update_minimum_bandwidth(self):
        vif_port = mock.Mock()
        vif_port.ofport = 'ofport'
        rule = mock.Mock()
        rule.min_kbps = 1500
        port = {'port_id': 'port_id', 'vif_port': vif_port,
                'physical_network': mock.ANY}
        with mock.patch.object(self.qos_driver.br_int,
                               'update_minimum_bandwidth_queue') \
                as mock_delete_minimum_bandwidth_queue, \
                mock.patch.object(self.qos_driver.agent_api,
                                  'request_phy_brs') as mock_request_phy_brs:
            phy_br = mock.Mock()
            phy_br.get_bridge_ports.return_value = ['port1', 'port2']
            mock_request_phy_brs.return_value = [phy_br]
            self.qos_driver.update_minimum_bandwidth(port, rule)
            mock_delete_minimum_bandwidth_queue.assert_called_once_with(
                'port_id', ['port1', 'port2'], 'ofport', 1500)

    # TODO(przszc): Update tests when dataplane enforcement is implemented for
    # minimum packet rate rule
    def test_create_minimum_packet_rate(self):
        try:
            port = {'port_id': 'p_id'}
            rule = mock.MagicMock(id='rule_id')
            self.qos_driver.create_minimum_packet_rate(port, rule)
        except Exception:
            self.fail('create_minimum_packet_rate failed')

    def test_update_minimum_packet_rate(self):
        try:
            port = {'port_id': 'p_id'}
            rule = mock.MagicMock(id='rule_id')
            self.qos_driver.update_minimum_packet_rate(port, rule)
        except Exception:
            self.fail('update_minimum_packet_rate failed')

    def test_delete_minimum_packet_rate(self):
        try:
            port = {'port_id': 'p_id'}
            self.qos_driver.delete_minimum_packet_rate(port)
        except Exception:
            self.fail('delete_minimum_packet_rate failed')

    def test_delete_minimum_packet_rate_ingress(self):
        try:
            port = {'port_id': 'p_id'}
            self.qos_driver.delete_minimum_packet_rate_ingress(port)
        except Exception:
            self.fail('delete_minimum_packet_rate_ingress failed')



# Copyright (c) 2018 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import itertools

from neutron_lib.db import standard_attr


def stdattrs_extended_resources(attributes):
    r_map = standard_attr.get_standard_attr_resource_model_map(
        include_resources=True,
        include_sub_resources=False)
    sr_map = standard_attr.get_standard_attr_resource_model_map(
        include_resources=False,
        include_sub_resources=True)
    return dict(itertools.chain(
        {r: attributes for r in r_map}.items(),
        {sr: {'parameters': attributes} for sr in sr_map}.items()
    ))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.objects.qos import binding
from neutron.objects import router
from neutron.tests.unit.objects import test_base
from neutron.tests.unit import testlib_api


class _QosPolicyBindingMixinTestCase(object):

    def test_get_bound_ids(self):
        [obj.create() for obj in self.objs]
        for obj in self.objs:
            obj_ids = obj.get_bound_ids(self.context, obj.policy_id)
            self.assertEqual(1, len(obj_ids))
            self.assertEqual(obj[obj.__class__._bound_model_id.name],
                             obj_ids[0])


class QosPolicyPortBindingObjectTestCase(test_base.BaseObjectIfaceTestCase):

    _test_class = binding.QosPolicyPortBinding


class QosPolicyPortBindingDbObjectTestCase(test_base.BaseDbObjectTestCase,
                                           testlib_api.SqlTestCase,
                                           _QosPolicyBindingMixinTestCase):

    _test_class = binding.QosPolicyPortBinding

    def setUp(self):
        super(QosPolicyPortBindingDbObjectTestCase, self).setUp()
        network_id = self._create_test_network_id()
        for db_obj in self.db_objs:
            self._create_test_qos_policy(id=db_obj['policy_id'])
            self._create_test_port(network_id=network_id,
                                   id=db_obj['port_id'])


class QosPolicyNetworkBindingObjectTestCase(test_base.BaseObjectIfaceTestCase):

    _test_class = binding.QosPolicyNetworkBinding


class QosPolicyNetworkBindingDbObjectTestCase(test_base.BaseDbObjectTestCase,
                                              testlib_api.SqlTestCase,
                                              _QosPolicyBindingMixinTestCase):

    _test_class = binding.QosPolicyNetworkBinding

    def setUp(self):
        super(QosPolicyNetworkBindingDbObjectTestCase, self).setUp()
        for db_obj in self.db_objs:
            self._create_test_qos_policy(id=db_obj['policy_id'])
            self._create_test_network(network_id=db_obj['network_id'])


class QosPolicyFloatingIPBindingObjectTestCase(
        test_base.BaseObjectIfaceTestCase):

    _test_class = binding.QosPolicyFloatingIPBinding


class QosPolicyFloatingIPBindingDbObjectTestCase(
        test_base.BaseDbObjectTestCase,
        testlib_api.SqlTestCase,
        _QosPolicyBindingMixinTestCase):

    _test_class = binding.QosPolicyFloatingIPBinding

    def setUp(self):
        super(QosPolicyFloatingIPBindingDbObjectTestCase, self).setUp()
        for db_obj in self.db_objs:
            self._create_test_qos_policy(id=db_obj['policy_id'])
            self._create_test_fip_id(fip_id=db_obj['fip_id'])


class QosPolicyRouterGatewayIPBindingObjectTestCase(
        test_base.BaseObjectIfaceTestCase):

    _test_class = binding.QosPolicyRouterGatewayIPBinding


class QosPolicyRouterGatewayIPBindingDbObjectTestCase(
        test_base.BaseDbObjectTestCase,
        testlib_api.SqlTestCase,
        _QosPolicyBindingMixinTestCase):

    _test_class = binding.QosPolicyRouterGatewayIPBinding

    def setUp(self):
        super(QosPolicyRouterGatewayIPBindingDbObjectTestCase, self).setUp()
        for db_obj in self.db_objs:
            self._create_test_qos_policy(id=db_obj['policy_id'])
            self._create_test_router_id(router_id=db_obj['router_id'])

    def test_get_routers_by_network_id(self):
        qos_policy_router_obj = self._create_test_qos_policy()
        qos_policy_net_obj = self._create_test_qos_policy()
        # External network 1, no QoS policy
        ext_network_id_1 = self._create_external_network_id()
        gw_port_id_1a = self._create_test_port_id(network_id=ext_network_id_1)
        gw_port_id_1b = self._create_test_port_id(network_id=ext_network_id_1)
        # External network 2, "qos_policy_network" assigned
        ext_network_id_2 = self._create_external_network_id(
            qos_policy_id=qos_policy_net_obj.id)
        gw_port_id_2a = self._create_test_port_id(network_id=ext_network_id_2)
        gw_port_id_2b = self._create_test_port_id(network_id=ext_network_id_2)

        # Router 1: no GW
        self._create_test_router_id(name='router1')

        # Router 2: GW assigned, no router QoS, not public network QoS
        router2 = self._create_test_router_id(name='router2')
        router2_obj = router.Router.get_object(self.context, id=router2)
        router2_obj.gw_port_id = gw_port_id_1a
        router2_obj.update()

        # Router 3: GW assigned, router QoS, not public network QoS
        router3 = self._create_test_router_id(name='router3')
        router3_obj = router.Router.get_object(self.context, id=router3)
        router3_obj.gw_port_id = gw_port_id_1b
        router3_obj.qos_policy_id = qos_policy_router_obj.id
        router3_obj.update()

        # Router 4: GW assigned, no router QoS, public network with QoS
        router4 = self._create_test_router_id(name='router4')
        router4_obj = router.Router.get_object(self.context, id=router4)
        router4_obj.gw_port_id = gw_port_id_2a
        router4_obj.update()

        # Router 5: GW assigned, router QoS, public network with QoS
        router5 = self._create_test_router_id(name='router5')
        router5_obj = router.Router.get_object(self.context, id=router5)
        router5_obj.gw_port_id = gw_port_id_2b
        router5_obj.qos_policy_id = qos_policy_router_obj.id
        router5_obj.update()

        # Check that only router3 and router5 have
        # "QosPolicyRouterGatewayIPBinding" related registers.
        qos_gw_binds = self._test_class.get_objects(self.context)
        self.assertEqual(2, len(qos_gw_binds))
        router_ids = [qos_gw_bind.router_id for qos_gw_bind in qos_gw_binds]
        self.assertEqual(sorted([router3, router5]), sorted(router_ids))

        result = self._test_class.get_routers_by_network_id(
            self.context, ext_network_id_1, policy_id=None)
        self.assertEqual([router2], [r.id for r in result])

        result = self._test_class.get_routers_by_network_id(
            self.context, ext_network_id_1, policy_id=qos_policy_router_obj.id)
        self.assertEqual([router3], [r.id for r in result])

        result = self._test_class.get_routers_by_network_id(
            self.context, ext_network_id_1, policy_id=qos_policy_net_obj.id)
        self.assertEqual([], [r.id for r in result])

        result = self._test_class.get_routers_by_network_id(
            self.context, ext_network_id_2, policy_id=None)
        self.assertEqual([router4], [r.id for r in result])

        result = self._test_class.get_routers_by_network_id(
            self.context, ext_network_id_2, policy_id=qos_policy_router_obj.id)
        self.assertEqual([router5], [r.id for r in result])

        result = self._test_class.get_routers_by_network_id(
            self.context, ext_network_id_2, policy_id=qos_policy_net_obj.id)
        self.assertEqual([], [r.id for r in result])



# Copyright (c) 2015 Mellanox Technologies, Ltd
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import collections

from neutron_lib.agent import l2_extension
from neutron_lib import constants
from neutron_lib.services.qos import constants as qos_consts
from oslo_concurrency import lockutils
from oslo_log import log as logging

from neutron.api.rpc.callbacks.consumer import registry
from neutron.api.rpc.callbacks import events
from neutron.api.rpc.callbacks import resources
from neutron.api.rpc.handlers import resources_rpc
from neutron import manager

LOG = logging.getLogger(__name__)


class QosAgentDriver(object, metaclass=abc.ABCMeta):
    """Defines stable abstract interface for QoS Agent Driver.

    QoS Agent driver defines the interface to be implemented by Agent
    for applying QoS Rules on a port.
    """

    # Each QoS driver should define the set of rule types that it supports, and
    # corresponding handlers that has the following names:
    #
    # create_<type>
    # update_<type>
    # delete_<type>
    #
    # where <type> is one of VALID_RULE_TYPES
    # There is exception from this rule for deletion of rules with
    # attribute direction set to ingress (e.g. bandwidth limit rule).
    # For deletion of such rule types delete handler has following name:
    # delete_<type>_ingress

    @abc.abstractmethod
    def initialize(self):
        """Perform QoS agent driver initialization.
        """

    def create(self, port, qos_policy):
        """Apply QoS rules on port for the first time.

        :param port: port object.
        :param qos_policy: the QoS policy to be applied on port.
        """
        self._handle_update_create_rules('create', port, qos_policy)

    def consume_api(self, agent_api):
        """Consume the AgentAPI instance from the QoSAgentExtension class

        This allows QosAgentDrivers to gain access to resources limited to the
        NeutronAgent when this method is overridden.

        :param agent_api: An instance of an agent specific API
        """

    def update(self, port, qos_policy):
        """Apply QoS rules on port.

        :param port: port object.
        :param qos_policy: the QoS policy to be applied on port.
        """
        self._handle_update_create_rules('update', port, qos_policy)

    def delete(self, port, qos_policy=None):
        """Remove QoS rules from port.

        :param port: port object.
        :param qos_policy: the QoS policy to be removed from port.
        """
        if qos_policy is None:
            for rule_type in self.SUPPORTED_RULES:
                self._handle_rule_delete(port, rule_type)
                if self._rule_type_has_ingress_direction(rule_type):
                    self._handle_rule_delete(port, rule_type, ingress=True)
        else:
            for rule in self._iterate_rules(qos_policy.rules):
                self._handle_rule_delete(
                    port, rule.rule_type,
                    ingress=self._rule_is_ingress_direction(rule))

    def _iterate_rules(self, rules):
        for rule in rules:
            rule_type = rule.rule_type
            if rule_type in self.SUPPORTED_RULES:
                yield rule
            else:
                LOG.warning('Unsupported QoS rule type for %(rule_id)s: '
                            '%(rule_type)s; skipping',
                            {'rule_id': rule.id, 'rule_type': rule_type})

    def _handle_rule_delete(self, port, rule_type, ingress=False):
        handler_name = "".join(("delete_", rule_type))
        if ingress:
            handler_name = "%s_%s" % (handler_name,
                                      constants.INGRESS_DIRECTION)
        handler = getattr(self, handler_name)
        handler(port)

    def _handle_update_create_rules(self, action, port, qos_policy):
        for rule in self._iterate_rules(qos_policy.rules):
            if rule.should_apply_to_port(port):
                handler_name = "".join((action, "_", rule.rule_type))
                handler = getattr(self, handler_name)
                handler(port, rule)
            else:
                LOG.debug("Port %(port)s excluded from QoS rule %(rule)s",
                          {'port': port, 'rule': rule.id})

    def _rule_type_has_ingress_direction(self, rule_type):
        supported_rule = self.SUPPORTED_RULES[rule_type]
        if qos_consts.DIRECTION not in supported_rule.keys():
            return False
        return (constants.INGRESS_DIRECTION in
                supported_rule[qos_consts.DIRECTION]['type:values'])

    def _rule_is_ingress_direction(self, rule):
        rule_direction = getattr(rule, "direction", constants.EGRESS_DIRECTION)
        return rule_direction == constants.INGRESS_DIRECTION


class PortPolicyMap(object):
    def __init__(self):
        # we cannot use a dict of sets here because port dicts are not hashable
        self.qos_policy_ports = collections.defaultdict(dict)
        self.known_policies = {}
        self.port_policies = {}

    def get_ports(self, policy):
        return self.qos_policy_ports[policy.id].values()

    def get_policy(self, policy_id):
        return self.known_policies.get(policy_id)

    def update_policy(self, policy):
        self.known_policies[policy.id] = policy

    def has_policy_changed(self, port, policy_id):
        return self.port_policies.get(port['port_id']) != policy_id

    def get_port_policy(self, port):
        policy_id = self.port_policies.get(port['port_id'])
        if policy_id:
            return self.get_policy(policy_id)

    def set_port_policy(self, port, policy):
        """Attach a port to policy and return any previous policy on port."""
        port_id = port['port_id']
        old_policy = self.get_port_policy(port)
        self.known_policies[policy.id] = policy
        self.port_policies[port_id] = policy.id
        self.qos_policy_ports[policy.id][port_id] = port
        if old_policy and old_policy.id != policy.id:
            del self.qos_policy_ports[old_policy.id][port_id]
        return old_policy

    def clean_by_port(self, port):
        """Detach port from policy and cleanup data we don't need anymore."""
        port_id = port['port_id']
        if port_id in self.port_policies:
            del self.port_policies[port_id]
            for qos_policy_id, port_dict in self.qos_policy_ports.items():
                if port_id in port_dict:
                    del port_dict[port_id]
                    if not port_dict:
                        self._clean_policy_info(qos_policy_id)
                    return
        LOG.debug("QoS extension did not have information on port %s",
                  port_id)

    def _clean_policy_info(self, qos_policy_id):
        del self.qos_policy_ports[qos_policy_id]
        del self.known_policies[qos_policy_id]


class QosAgentExtension(l2_extension.L2AgentExtension):
    SUPPORTED_RESOURCE_TYPES = [resources.QOS_POLICY]

    def initialize(self, connection, driver_type):
        """Initialize agent extension."""

        self.resource_rpc = resources_rpc.ResourcesPullRpcApi()
        self.qos_driver = manager.NeutronManager.load_class_for_provider(
            'neutron.qos.agent_drivers', driver_type)()
        self.qos_driver.consume_api(self.agent_api)
        self.qos_driver.initialize()

        self.policy_map = PortPolicyMap()

        self._register_rpc_consumers(connection)

    def consume_api(self, agent_api):
        """Allows an extension to gain access to resources internal to the
           neutron agent and otherwise unavailable to the extension.
        """
        self.agent_api = agent_api

    def _register_rpc_consumers(self, connection):
        """Allows an extension to receive notifications of updates made to
           items of interest.
        """
        endpoints = [resources_rpc.ResourcesPushRpcCallback()]
        for resource_type in self.SUPPORTED_RESOURCE_TYPES:
            # We assume that the neutron server always broadcasts the latest
            # version known to the agent
            registry.register(self._handle_notification, resource_type)
            topic = resources_rpc.resource_type_versioned_topic(resource_type)
            connection.create_consumer(topic, endpoints, fanout=True)

    @lockutils.synchronized('qos-port')
    def _handle_notification(self, context, resource_type,
                             qos_policies, event_type):
        # server does not allow to remove a policy that is attached to any
        # port, so we ignore DELETED events. Also, if we receive a CREATED
        # event for a policy, it means that there are no ports so far that are
        # attached to it. That's why we are interested in UPDATED events only
        if event_type == events.UPDATED:
            for qos_policy in qos_policies:
                self._process_update_policy(qos_policy)

    @lockutils.synchronized('qos-port')
    def handle_port(self, context, port):
        """Handle agent QoS extension for port.

        This method applies a new policy to a port using the QoS driver.
        Update events are handled in _handle_notification.
        """
        port_id = port['port_id']
        port_qos_policy_id = port.get(qos_consts.QOS_POLICY_ID)
        qos_network_policy_id = port.get(qos_consts.QOS_NETWORK_POLICY_ID)
        qos_policy_id = port_qos_policy_id or qos_network_policy_id
        if qos_policy_id is None:
            self._process_reset_port(port)
            return

        if not self.policy_map.has_policy_changed(port, qos_policy_id):
            return

        qos_policy = self.policy_map.get_policy(
            qos_policy_id) or self.resource_rpc.pull(
                context, resources.QOS_POLICY, qos_policy_id)
        if qos_policy is None:
            LOG.info("QoS policy %(qos_policy_id)s applied to port "
                     "%(port_id)s is not available on server, "
                     "it has been deleted. Skipping.",
                     {'qos_policy_id': qos_policy_id, 'port_id': port_id})
            self._process_reset_port(port)
        else:
            old_qos_policy = self.policy_map.set_port_policy(port, qos_policy)
            # Before applying the new rules, the old rules should be cleared,
            # even if the old_qos_policy is None,
            # to avoid the data being out of sync before the l2-agent starts.
            self.qos_driver.delete(port, old_qos_policy)
            if qos_policy.rules:
                self.qos_driver.create(port, qos_policy)

    def delete_port(self, context, port):
        self._process_reset_port(port)

    def _policy_rules_modified(self, old_policy, policy):
        return not (len(old_policy.rules) == len(policy.rules) and
                    all(i in old_policy.rules for i in policy.rules))

    def _process_update_policy(self, qos_policy):
        old_qos_policy = self.policy_map.get_policy(qos_policy.id)
        if old_qos_policy:
            if self._policy_rules_modified(old_qos_policy, qos_policy):
                for port in self.policy_map.get_ports(qos_policy):
                    # NOTE(QoS): for now, just reflush the rules on the port.
                    #            Later, we may want to apply the difference
                    #            between the old and new rule lists.
                    self.qos_driver.delete(port, old_qos_policy)
                    self.qos_driver.update(port, qos_policy)
            self.policy_map.update_policy(qos_policy)

    def _process_reset_port(self, port):
        self.policy_map.clean_by_port(port)
        self.qos_driver.delete(port)



# Copyright 2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections

from neutron_lib.api.definitions import portbindings
from neutron_lib.db import api as db_api
from neutron_lib.plugins import directory
from neutron_lib import rpc as n_rpc
from neutron_lib.services.trunk import constants as trunk_consts
from oslo_log import helpers as log_helpers
from oslo_log import log as logging
import oslo_messaging
from sqlalchemy.orm import exc

from neutron.api.rpc.callbacks import events
from neutron.api.rpc.callbacks.producer import registry
from neutron.api.rpc.callbacks import resources
from neutron.api.rpc.handlers import resources_rpc
from neutron.objects import trunk as trunk_objects
from neutron.services.trunk import exceptions as trunk_exc
from neutron.services.trunk.rpc import constants

LOG = logging.getLogger(__name__)

# This module contains stub (client-side) and skeleton (server-side)
# proxy code that executes in the Neutron server process space. This
# is needed if any of the trunk service plugin drivers has a remote
# component (e.g. agent), that needs to communicate with the Neutron
# Server.

# The Server side exposes the following remote methods:
#
# - lookup method to retrieve trunk details: used by the agent to learn
#   about the trunk.
# - update methods for trunk and its subports: used by the agent to
#   inform the server about local trunk status changes.
#
# For agent-side stub and skeleton proxy code, please look at agent.py


def trunk_by_port_provider(resource, port_id, context, **kwargs):
    """Provider callback to supply trunk information by parent port."""
    return trunk_objects.Trunk.get_object(context, port_id=port_id)


class TrunkSkeleton(object):
    """Skeleton proxy code for agent->server communication."""

    # API version history:
    # 1.0 Initial version
    target = oslo_messaging.Target(version='1.0',
                                   namespace=constants.TRUNK_BASE_NAMESPACE)

    _core_plugin = None

    def __init__(self):
        # Used to provide trunk lookups for the agent.
        registry.provide(trunk_by_port_provider, resources.TRUNK)
        self._connection = n_rpc.Connection()
        self._connection.create_consumer(
            constants.TRUNK_BASE_TOPIC, [self], fanout=False)
        self._connection.consume_in_threads()

    @property
    def core_plugin(self):
        if not self._core_plugin:
            self._core_plugin = directory.get_plugin()
        return self._core_plugin

    @log_helpers.log_method_call
    def update_subport_bindings(self, context, subports):
        """Update subport bindings to match trunk host binding."""
        el = context.elevated()
        ports_by_trunk_id = collections.defaultdict(list)
        updated_ports = collections.defaultdict(list)

        for s in subports:
            ports_by_trunk_id[s['trunk_id']].append(s['port_id'])
        for trunk_id, subport_ids in ports_by_trunk_id.items():
            trunk = trunk_objects.Trunk.get_object(el, id=trunk_id)
            if not trunk:
                LOG.debug("Trunk not found. id: %s", trunk_id)
                continue

            trunk_updated_ports = self._process_trunk_subport_bindings(
                el, trunk, subport_ids)
            updated_ports[trunk.id].extend(trunk_updated_ports)

        return updated_ports

    def _safe_update_trunk(self, trunk, **kwargs):
        for try_cnt in range(db_api.MAX_RETRIES):
            try:
                trunk.update(**kwargs)
                break
            except exc.StaleDataError as e:
                if try_cnt < db_api.MAX_RETRIES - 1:
                    LOG.debug("Got StaleDataError exception: %s", e)
                    continue
                # re-raise when all tries failed
                raise

    def update_trunk_status(self, context, trunk_id, status):
        """Update the trunk status to reflect outcome of data plane wiring."""
        with db_api.CONTEXT_WRITER.using(context):
            trunk = trunk_objects.Trunk.get_object(context, id=trunk_id)
            if trunk:
                self._safe_update_trunk(trunk, status=status)

    def _process_trunk_subport_bindings(self, context, trunk, port_ids):
        """Process port bindings for subports on the given trunk."""
        updated_ports = []
        trunk_port_id = trunk.port_id
        trunk_port = self.core_plugin.get_port(context, trunk_port_id)
        trunk_host = trunk_port.get(portbindings.HOST_ID)
        migrating_to_host = trunk_port.get(
            portbindings.PROFILE, {}).get('migrating_to')
        if migrating_to_host and trunk_host != migrating_to_host:
            # Trunk is migrating now, so lets update host of the subports
            # to the new host already
            trunk_host = migrating_to_host

        # NOTE(status_police) Set the trunk in BUILD state before
        # processing subport bindings. The trunk will stay in BUILD
        # state until an attempt has been made to bind all subports
        # passed here and the agent acknowledges the operation was
        # successful.
        self._safe_update_trunk(
            trunk, status=trunk_consts.TRUNK_BUILD_STATUS)

        for port_id in port_ids:
            try:
                updated_port = self._handle_port_binding(context, port_id,
                                                         trunk, trunk_host)
                # NOTE(fitoduarte): consider trimming down the content
                # of the port data structure.
                updated_ports.append(updated_port)
            except trunk_exc.SubPortBindingError as e:
                LOG.error("Failed to bind subport: %s", e)

                # NOTE(status_police) The subport binding has failed in a
                # manner in which we cannot proceed and the user must take
                # action to bring the trunk back to a sane state.
                self._safe_update_trunk(
                    trunk, status=trunk_consts.TRUNK_ERROR_STATUS)
                return []
            except Exception as e:
                msg = ("Failed to bind subport port %(port)s on trunk "
                       "%(trunk)s: %(exc)s")
                LOG.error(msg, {'port': port_id, 'trunk': trunk.id, 'exc': e})

        if len(port_ids) != len(updated_ports):
            self._safe_update_trunk(
                trunk, status=trunk_consts.TRUNK_DEGRADED_STATUS)

        return updated_ports

    def _handle_port_binding(self, context, port_id, trunk, trunk_host):
        """Bind the given port to the given host.

           :param context: The context to use for the operation
           :param port_id: The UUID of the port to be bound
           :param trunk: The trunk that the given port belongs to
           :param trunk_host: The host to bind the given port to
        """
        port = self.core_plugin.update_port(
            context, port_id,
            {'port': {portbindings.HOST_ID: trunk_host,
                      'device_owner': trunk_consts.TRUNK_SUBPORT_OWNER}})
        vif_type = port.get(portbindings.VIF_TYPE)
        if vif_type == portbindings.VIF_TYPE_BINDING_FAILED:
            raise trunk_exc.SubPortBindingError(port_id=port_id,
                                                trunk_id=trunk.id)
        return port


class TrunkStub(object):
    """Stub proxy code for server->agent communication."""

    def __init__(self):
        self._resource_rpc = resources_rpc.ResourcesPushRpcApi()

    @log_helpers.log_method_call
    def trunk_created(self, context, trunk):
        """Tell the agent about a trunk being created."""
        self._resource_rpc.push(context, [trunk], events.CREATED)

    @log_helpers.log_method_call
    def trunk_deleted(self, context, trunk):
        """Tell the agent about a trunk being deleted."""
        self._resource_rpc.push(context, [trunk], events.DELETED)

    @log_helpers.log_method_call
    def subports_added(self, context, subports):
        """Tell the agent about new subports to add."""
        self._resource_rpc.push(context, subports, events.CREATED)

    @log_helpers.log_method_call
    def subports_deleted(self, context, subports):
        """Tell the agent about existing subports to remove."""
        self._resource_rpc.push(context, subports, events.DELETED)



# Copyright 2016 Hewlett Packard Enterprise Development Company LP
#
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from neutron.services.trunk.drivers.linuxbridge import driver as lxb_driver
from neutron.services.trunk.drivers.openvswitch import driver as ovs_driver


def register():
    """Load in-tree drivers for the service plugin."""
    # Enable the trunk plugin to work with ML2/OVS. Support for other
    # drivers can be added similarly by executing the registration
    # code at the time of plugin/mech driver initialization. There should
    # be at least one compatible driver enabled in the deployment for trunk
    # setup to be successful. The plugin fails to initialize if no compatible
    # driver is found in the deployment.
    lxb_driver.register()
    ovs_driver.register()



# Copyright 2015 Mirantis Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import multiprocessing
import os
import queue
import signal
import socket
import time
import traceback
from unittest import mock

import httplib2
from neutron_lib import worker as neutron_worker
from oslo_config import cfg
from oslo_log import log
import psutil

from neutron.common import utils
from neutron import manager
from neutron import service
from neutron.tests import base as tests_base
from neutron.tests.functional import base
from neutron import wsgi


LOG = log.getLogger(__name__)

CONF = cfg.CONF

# Those messages will be written to temporary file each time
# start/reset methods are called.
FAKE_START_MSG = 'start'
FAKE_RESET_MSG = 'reset'

TARGET_PLUGIN = 'neutron.plugins.ml2.plugin.Ml2Plugin'


class TestNeutronServer(base.BaseLoggingTestCase):
    def setUp(self):
        super(TestNeutronServer, self).setUp()
        self.service_pid = None
        self.workers = None
        self._mp_queue = multiprocessing.Queue()
        self.health_checker = self._check_active
        self.pipein, self.pipeout = os.pipe()
        self.addCleanup(self._destroy_workers)

    def _destroy_workers(self):
        if self.service_pid:
            # Make sure all processes are stopped
            os.kill(self.service_pid, signal.SIGKILL)

    def _start_server(self, callback, workers, processes_queue=None):
        """Run a given service.

        :param callback: callback that will start the required service
        :param workers: number of service workers
        :returns: list of spawned workers' pids
        """

        self.workers = workers

        # Fork a new process in which server will be started
        pid = os.fork()
        if pid == 0:
            status = 0
            try:
                callback(workers)
            except SystemExit as exc:
                status = exc.code
            except BaseException:
                traceback.print_exc()
                status = 2

            # Really exit
            os._exit(status)

        self.service_pid = pid

        # If number of workers is 1 it is assumed that we run
        # a service in the current process.
        if self.workers > 1:
            workers = self._get_workers(10, processes_queue=processes_queue)
            self.assertEqual(len(workers), self.workers)
            return workers

        # Wait for a service to start.
        utils.wait_until_true(self.health_checker, timeout=10, sleep=0.1,
                              exception=RuntimeError(
                                  "Failed to start service."))

        return [self.service_pid]

    def _get_workers(self, timeout, processes_queue=None):
        """Get the list of processes in which WSGI server is running."""
        def safe_ppid(proc):
            try:
                return proc.ppid()
            except psutil.NoSuchProcess:
                return None

        def get_workers_pid():
            if self.workers > 1:
                return [proc.pid for proc in psutil.process_iter()
                        if safe_ppid(proc) == self.service_pid]
            else:
                return [proc.pid for proc in psutil.process_iter()
                        if proc.pid == self.service_pid]

        exception = RuntimeError('Failed to start %d workers.' % self.workers)

        if processes_queue:
            try:
                return processes_queue.get(timeout=timeout)
            except queue.Empty:
                raise exception

        # Wait at most 10 seconds to spawn workers
        condition = lambda: self.workers == len(get_workers_pid())
        utils.wait_until_true(condition, timeout=timeout, sleep=0.1,
                              exception=exception)
        return get_workers_pid()

    def _check_active(self):
        """Dummy service activity check."""
        time.sleep(5)
        return True

    def _fake_start(self):
        self._mp_queue.put(FAKE_START_MSG)

    def _fake_reset(self):
        self._mp_queue.put(FAKE_RESET_MSG)

    def _test_restart_service_on_sighup(self, service, workers=1,
                                        processes_queue=None):
        """Test that a service correctly (re)starts on receiving SIGHUP.

        1. Start a service with a given number of workers.
        2. Send SIGHUP to the service.
        3. Wait for workers (if any) to (re)start.
        """

        self._start_server(callback=service, workers=workers,
                           processes_queue=processes_queue)
        os.kill(self.service_pid, signal.SIGHUP)

        # After sending SIGHUP it is expected that there will be as many
        # FAKE_RESET_MSG as number of workers + one additional for main
        # process
        expected_msg = (
            FAKE_START_MSG * workers + FAKE_RESET_MSG * (workers + 1))

        # Wait for temp file to be created and its size reaching the expected
        # value
        expected_size = len(expected_msg)
        ret_msg = ''

        def is_ret_buffer_ok():
            nonlocal ret_msg
            LOG.debug('Checking returned buffer size')
            while not self._mp_queue.empty():
                ret_msg += self._mp_queue.get()
            LOG.debug('Size of buffer is %s. Expected size: %s',
                      len(ret_msg), expected_size)
            return len(ret_msg) == expected_size

        try:
            utils.wait_until_true(is_ret_buffer_ok, timeout=5, sleep=1)
        except utils.WaitTimeout:
            raise RuntimeError('Expected buffer size: %s, current size: %s' %
                               (len(ret_msg), expected_size))

        # Verify that start has been called twice for each worker (one for
        # initial start, and the second one on SIGHUP after children were
        # terminated).
        self.assertEqual(expected_msg, ret_msg)


class TestWsgiServer(TestNeutronServer):
    """Tests for neutron.wsgi.Server."""

    def setUp(self):
        super(TestWsgiServer, self).setUp()
        self.health_checker = self._check_active
        self.port = None

    @staticmethod
    def application(environ, start_response):
        """A primitive test application."""

        response_body = 'Response'
        status = '200 OK'
        response_headers = [('Content-Type', 'text/plain'),
                            ('Content-Length', str(len(response_body)))]
        start_response(status, response_headers)
        return [response_body]

    def _check_active(self):
        """Check a wsgi service is active by making a GET request."""
        port = int(os.read(self.pipein, 5))
        conn = httplib2.HTTPConnectionWithTimeout("localhost", port)
        try:
            conn.request("GET", "/")
            resp = conn.getresponse()
            return resp.status == 200
        except socket.error:
            return False

    def _run_wsgi(self, workers=1):
        """Start WSGI server with a test application."""

        # Mock start method to check that children are started again on
        # receiving SIGHUP.
        with mock.patch("neutron.wsgi.WorkerService.start") as start_method,\
                mock.patch("neutron.wsgi.WorkerService.reset") as reset_method:
            start_method.side_effect = self._fake_start
            reset_method.side_effect = self._fake_reset

            server = wsgi.Server("Test")
            server.start(self.application, 0, "0.0.0.0",
                         workers=workers)

            # Memorize a port that was chosen for the service
            self.port = server.port
            os.write(self.pipeout, bytes(str(self.port), 'utf-8'))

            server.wait()

    @tests_base.unstable_test('bug 1930367')
    def test_restart_wsgi_on_sighup_multiple_workers(self):
        self._test_restart_service_on_sighup(service=self._run_wsgi,
                                             workers=2)


class TestRPCServer(TestNeutronServer):
    """Tests for neutron RPC server."""

    def setUp(self):
        super(TestRPCServer, self).setUp()
        self.setup_coreplugin('ml2', load_plugins=False)
        self._plugin_patcher = mock.patch(TARGET_PLUGIN, autospec=True)
        self.plugin = self._plugin_patcher.start()
        self.plugin.return_value.rpc_workers_supported = True
        self._processes_queue = multiprocessing.Queue()

    def _serve_rpc(self, workers=1):
        """Start RPC server with a given number of workers."""

        # Mock start method to check that children are started again on
        # receiving SIGHUP.
        with mock.patch("neutron.service.RpcWorker.start") as start_method,\
                mock.patch(
                    "neutron.service.RpcWorker.reset") as reset_method,\
                mock.patch(
                    "neutron_lib.plugins.directory.get_plugin") as get_plugin:
            start_method.side_effect = self._fake_start
            reset_method.side_effect = self._fake_reset
            get_plugin.return_value = self.plugin

            CONF.set_override("rpc_workers", workers)
            # not interested in state report workers specifically
            CONF.set_override("rpc_state_report_workers", 0)

            rpc_workers_launcher = service.start_rpc_workers()
            self._processes_queue.put(list(rpc_workers_launcher.children))
            rpc_workers_launcher.wait()

    def test_restart_rpc_on_sighup_multiple_workers(self):
        self._test_restart_service_on_sighup(
            service=self._serve_rpc, workers=2,
            processes_queue=self._processes_queue)


class TestPluginWorker(TestNeutronServer):
    """Ensure that a plugin returning Workers spawns workers"""

    def setUp(self):
        super(TestPluginWorker, self).setUp()
        self.setup_coreplugin('ml2', load_plugins=False)
        self._plugin_patcher = mock.patch(TARGET_PLUGIN, autospec=True)
        self.plugin = self._plugin_patcher.start()
        manager.init()

    def _start_plugin(self, workers=1):
        with mock.patch('neutron_lib.plugins.directory.get_plugin') as gp:
            gp.return_value = self.plugin
            plugin_workers_launcher = service.start_plugins_workers()
            plugin_workers_launcher.wait()

    def test_start(self):
        class FakeWorker(neutron_worker.BaseWorker):
            def start(self):
                pass

            def wait(self):
                pass

            def stop(self):
                pass

            def reset(self):
                pass

        # Make both ABC happy and ensure 'self' is correct
        FakeWorker.start = self._fake_start
        FakeWorker.reset = self._fake_reset
        workers = [FakeWorker()]
        self.plugin.return_value.get_workers.return_value = workers
        self._test_restart_service_on_sighup(service=self._start_plugin,
                                             workers=len(workers))



# Copyright 2019 Ericsson
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from keystoneauth1 import exceptions as ks_exc
from neutron_lib.agent import constants as agent_const
from oslo_log import log as logging

from neutron.services.placement_report import plugin
from neutron.tests.unit.plugins.ml2.drivers import mechanism_test
from neutron.tests.unit.plugins.ml2 import test_plugin

LOG = logging.getLogger(__name__)


class PlacementReportPluginTestCases(test_plugin.Ml2PluginV2TestCase):

    _mechanism_drivers = ['logger', 'test_with_agent']

    def setUp(self):
        super(PlacementReportPluginTestCases, self).setUp()
        self.service_plugin = plugin.PlacementReportPlugin()

    def test__get_rp_by_name_found(self):
        with mock.patch.object(
                self.service_plugin._placement_client,
                'list_resource_providers',
                return_value={'resource_providers': ['fake_rp']}):
            rp = self.service_plugin._get_rp_by_name('whatever')
        self.assertEqual('fake_rp', rp)

    def test__get_rp_by_name_not_found(self):
        with mock.patch.object(
                self.service_plugin._placement_client,
                'list_resource_providers',
                return_value={'resource_providers': []}):
            self.assertRaises(
                IndexError, self.service_plugin._get_rp_by_name, 'no_such_rp')

    def test_no_sync_for_rp_name_not_found(self):
        # looking all good
        agent = {
            'agent_type': 'test_mechanism_driver_agent',
            'configurations': {
                'resource_provider_bandwidths': {'some iface': ''},
            },
            'host': 'fake host',
        }
        agent_db = mock.Mock()

        with mock.patch.object(
                self.service_plugin._placement_client,
                'list_resource_providers',
                return_value={'resource_providers': []}), \
            mock.patch.object(
                self.service_plugin._batch_notifier,
                'queue_event') as mock_queue_event:

            self.service_plugin._sync_placement_state(agent, agent_db)

            self.assertFalse(agent_db.resources_synced)
            agent_db.update.assert_called_with()
            mock_queue_event.assert_not_called()

    def test_no_sync_for_placement_gone(self):
        # looking all good
        agent = {
            'agent_type': 'test_mechanism_driver_agent',
            'configurations': {
                'resource_provider_bandwidths': {'some iface': ''}},
            'host': 'fake host',
        }
        agent_db = mock.Mock()

        with mock.patch.object(
                self.service_plugin._placement_client,
                'list_resource_providers',
                side_effect=ks_exc.HttpError), \
            mock.patch.object(
                self.service_plugin._batch_notifier,
                'queue_event') as mock_queue_event:

            self.service_plugin._sync_placement_state(agent, agent_db)

            self.assertFalse(agent_db.resources_synced)
            agent_db.update.assert_called_with()
            mock_queue_event.assert_not_called()

    def test_no_sync_for_unsupported_agent_type(self):
        payload = mock.Mock(
            # looking all good, but agent type not supported
            desired_state={
                'agent_type': 'unsupported agent type',
                'configurations': {'resource_provider_bandwidths': {}},
                'host': 'fake host',
            })

        with mock.patch.object(self.service_plugin._core_plugin,
                '_get_agent_by_type_and_host') as mock_get_agent, \
            mock.patch.object(self.service_plugin,
                '_sync_placement_state') as mock_sync:

            self.service_plugin.handle_placement_config(
                mock.ANY, mock.ANY, mock.ANY, payload)

            mock_get_agent.assert_not_called()
            mock_sync.assert_not_called()

    def test_no_sync_without_resource_info(self):
        payload = mock.Mock(
            # looking all good, but 'configurations' has no
            # 'resource_provider_bandwidths'
            desired_state={
                'agent_type': 'test_mechanism_driver_agent',
                'configurations': {},
                'host': 'fake host',
            })

        with mock.patch.object(self.service_plugin._core_plugin,
                '_get_agent_by_type_and_host') as mock_get_agent, \
            mock.patch.object(self.service_plugin,
                '_sync_placement_state') as mock_sync:

            self.service_plugin.handle_placement_config(
                mock.ANY, mock.ANY, mock.ANY, payload)

            mock_get_agent.assert_not_called()
            mock_sync.assert_not_called()

    def test_sync_if_agent_is_new(self):
        payload = mock.Mock(
            desired_state={
                'agent_type': 'test_mechanism_driver_agent',
                'configurations': {'resource_provider_bandwidths': {}},
                'host': 'fake host',
            },
            metadata={
                'status': agent_const.AGENT_NEW,
            },
        )

        with mock.patch.object(self.service_plugin._core_plugin,
                '_get_agent_by_type_and_host') as mock_get_agent, \
            mock.patch.object(self.service_plugin,
                '_sync_placement_state') as mock_sync:

            self.service_plugin.handle_placement_config(
                mock.ANY, mock.ANY, mock.ANY, payload)

            self.assertEqual(1, mock_get_agent.call_count)
            self.assertEqual(1, mock_sync.call_count)

    def test_sync_if_agent_is_restarted(self):
        payload = mock.Mock(
            desired_state={
                'agent_type': 'test_mechanism_driver_agent',
                'configurations': {'resource_provider_bandwidths': {}},
                'host': 'fake host',
                'start_flag': True,
            },
        )

        with mock.patch.object(self.service_plugin._core_plugin,
                '_get_agent_by_type_and_host') as mock_get_agent, \
            mock.patch.object(self.service_plugin,
                '_sync_placement_state') as mock_sync:

            self.service_plugin.handle_placement_config(
                mock.ANY, mock.ANY, mock.ANY, payload)

            self.assertEqual(1, mock_get_agent.call_count)
            self.assertEqual(1, mock_sync.call_count)

    def test_sync_after_transient_error(self):
        payload = mock.Mock(
            desired_state={
                'agent_type': 'test_mechanism_driver_agent',
                'configurations': {'resource_provider_bandwidths': {}},
                'host': 'fake host',
            },
        )

        with mock.patch.object(self.service_plugin._core_plugin,
                '_get_agent_by_type_and_host',
                return_value={'resources_synced': False}) as mock_get_agent, \
            mock.patch.object(self.service_plugin,
                '_sync_placement_state') as mock_sync:

            self.service_plugin.handle_placement_config(
                mock.ANY, mock.ANY, mock.ANY, payload)

            self.assertEqual(1, mock_get_agent.call_count)
            self.assertEqual(1, mock_sync.call_count)

    def test__sync_placement_state_legacy(self):
        agent = {
            'agent_type': 'test_mechanism_driver_agent',
            'configurations': {
                'resource_provider_bandwidths': {},
                'resource_provider_inventory_defaults': {},
            },
            'host': 'fake host',
        }
        agent_db = mock.Mock()

        with mock.patch.object(self.service_plugin._batch_notifier,
                'queue_event') as mock_queue_event, \
            mock.patch.object(self.service_plugin._placement_client,
               'list_resource_providers',
               return_value={'resource_providers': [{'uuid': 'fake uuid'}]}):

            self.service_plugin._sync_placement_state(agent, agent_db)

            self.assertEqual(1, mock_queue_event.call_count)

    def test__sync_placement_state_rp_hypervisors(self):
        agent = {
            'agent_type': 'test_mechanism_driver_agent',
            'configurations': {
                'resource_provider_bandwidths': {},
                'resource_provider_inventory_defaults': {},
                'resource_provider_hypervisors': {'eth0': 'hypervisor0'},
            },
            'host': 'fake host',
        }
        agent_db = mock.Mock()

        with mock.patch.object(self.service_plugin._batch_notifier,
                'queue_event') as mock_queue_event, \
            mock.patch.object(self.service_plugin._placement_client,
               'list_resource_providers',
               return_value={'resource_providers': [
                   {'uuid': 'fake uuid'}]}) as mock_list_rps:

            self.service_plugin._sync_placement_state(agent, agent_db)

            self.assertEqual(1, mock_queue_event.call_count)
            mock_list_rps.assert_called_once_with(name='hypervisor0')

    def test__sync_placement_state_rp_pkt_processing_with_direction(self):
        agent = {
            'agent_type': 'test_mechanism_driver_agent',
            'configurations': {
                'resource_provider_bandwidths': {},
                'resource_provider_inventory_defaults': {},
                'resource_provider_packet_processing_with_direction': {
                    'fake host': {'egress': 1, 'ingress': 2}
                },
                'resource_provider_packet_processing_inventory_defaults': {
                    'allocation_ratio': 1, 'min_unit': 1, 'step_size': 1
                },
            },
            'host': 'fake host',
        }
        agent_db = mock.Mock()
        mock_state = mock.Mock(return_value=[])

        with mock.patch.object(self.service_plugin._batch_notifier,
                'queue_event') as mock_queue_event, \
            mock.patch('neutron.agent.common.placement_report.PlacementState',
                return_value=mock_state) as mock_placement_state:

            self.service_plugin._sync_placement_state(agent, agent_db)

            self.assertEqual(1, mock_queue_event.call_count)
            mock_placement_state.assert_called_once_with(
                rp_pkt_processing={'fake host': {'egress': 1, 'ingress': 2}},
                rp_pkt_processing_inventory_defaults={
                    'allocation_ratio': 1, 'min_unit': 1, 'step_size': 1},
                rp_bandwidths=mock.ANY,
                rp_inventory_defaults=mock.ANY,
                driver_uuid_namespace=mock.ANY,
                agent_type=mock.ANY,
                hypervisor_rps=mock.ANY,
                device_mappings=mock.ANY,
                supported_vnic_types=mock.ANY,
                client=mock.ANY)
            mock_state.deferred_sync.assert_called_once()

    def test__sync_placement_state_rp_pkt_processing_without_direction(self):
        agent = {
            'agent_type': 'test_mechanism_driver_agent',
            'configurations': {
                'resource_provider_bandwidths': {},
                'resource_provider_inventory_defaults': {},
                'resource_provider_packet_processing_without_direction': {
                    'fake host': {'any': 1}
                },
                'resource_provider_packet_processing_inventory_defaults': {
                    'allocation_ratio': 1, 'min_unit': 1, 'step_size': 1
                },
            },
            'host': 'fake host',
        }
        agent_db = mock.Mock()
        mock_state = mock.Mock(return_value=[])

        with mock.patch.object(self.service_plugin._batch_notifier,
                'queue_event') as mock_queue_event, \
            mock.patch('neutron.agent.common.placement_report.PlacementState',
                return_value=mock_state) as mock_placement_state:

            self.service_plugin._sync_placement_state(agent, agent_db)

            self.assertEqual(1, mock_queue_event.call_count)
            mock_placement_state.assert_called_once_with(
                rp_pkt_processing={'fake host': {'any': 1}},
                rp_pkt_processing_inventory_defaults={
                    'allocation_ratio': 1, 'min_unit': 1, 'step_size': 1},
                rp_bandwidths=mock.ANY,
                rp_inventory_defaults=mock.ANY,
                driver_uuid_namespace=mock.ANY,
                agent_type=mock.ANY,
                hypervisor_rps=mock.ANY,
                device_mappings=mock.ANY,
                supported_vnic_types=mock.ANY,
                client=mock.ANY)
            mock_state.deferred_sync.assert_called_once()


class PlacementReporterAgentsTestCases(test_plugin.Ml2PluginV2TestCase):

    _mechanism_drivers = ['logger', 'test_with_agent']

    def test_supported_agent_types(self):
        self.agents = plugin.PlacementReporterAgents(ml2_plugin=self.plugin)
        self.assertEqual(
            ['test_mechanism_driver_agent'],
            self.agents.supported_agent_types)

    def test_mechanism_driver_by_agent_type_found(self):
        self.agents = plugin.PlacementReporterAgents(ml2_plugin=self.plugin)
        mech_driver = self.agents.mechanism_driver_by_agent_type(
            'test_mechanism_driver_agent')
        self.assertIsInstance(mech_driver, mechanism_test.TestMechanismDriver)

    def test_mechanism_driver_by_agent_type_not_found(self):
        self.agents = plugin.PlacementReporterAgents(ml2_plugin=self.plugin)
        self.assertRaises(
            Exception,  # noqa
            self.agents.mechanism_driver_by_agent_type,
            'agent_not_belonging_to_any_mechanism_driver')



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from oslo_utils import importutils

from neutron.tests import base
from neutron.tests.unit.common import moved_globals_target as new_mod


def module_path(code):
    return 'neutron.tests.unit.common.moved_globals_' + code


def import_code(code):
    return importutils.import_module(module_path(code))


def expect_moved(code, name, new_name=None):
    old_path = '.'.join([module_path(code), name])
    new_path = '.'.join([new_mod.__name__, new_name or name])
    message = 'moved to ' + new_path
    return old_path, message


def expect_renamed(code, old_name, new_name):
    old_path = '.'.join([module_path(code), old_name])
    new_path = '.'.join([module_path(code), new_name])
    message = 'renamed to ' + new_path
    return old_path, message


class TestMovedGlobals(base.BaseTestCase):

    def test_moved_global(self):
        code = 'code1'
        old_mod = import_code(code)
        with mock.patch('debtcollector.deprecate') as dc:
            self.assertEqual(new_mod.a, old_mod.a)
            old_path, msg = expect_moved(code, 'a')
            dc.assert_called_once_with(old_path, message=msg, stacklevel=4)

    def test_moved_global_no_attr(self):
        mod = import_code('code1')
        self.assertRaises(AttributeError, lambda: mod.NO_SUCH_ATTRIBUTE)

    def test_renamed_global(self):
        code = 'code1'
        mod = import_code(code)
        with mock.patch('debtcollector.deprecate') as dc:
            self.assertEqual(mod.d, mod.c)
            old_path, msg = expect_renamed(code, 'c', 'd')
            dc.assert_called_once_with(old_path, message=msg, stacklevel=4)

    def test_moved_global_renamed(self):
        code = 'code1'
        old_mod = import_code(code)
        with mock.patch('debtcollector.deprecate') as dc:
            self.assertEqual(new_mod.f, old_mod.e)
            old_path, msg = expect_moved(code, 'e', new_name='f')
            dc.assert_called_once_with(old_path, message=msg, stacklevel=4)

    def test_set_unmoved_global(self):
        mod = import_code('code1')
        mod.d = 'dibatag'
        self.assertEqual('dibatag', mod.d)

    def test_set_new_global(self):
        mod = import_code('code1')
        mod.n = 'nyala'
        self.assertEqual('nyala', mod.n)

    def test_delete_unmoved_global(self):
        mod = import_code('code1')
        self.assertEqual('gelada', mod.g)

        def delete_g():
            del mod.g

        delete_g()
        self.assertRaises(AttributeError, lambda: mod.g)
        self.assertRaises(AttributeError, delete_g)

    def test_not_last_line(self):
        self.assertRaises(SystemExit, import_code, 'code2')



# Copyright (c) 2022 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils

from neutron.conf import quota as quota_conf
from neutron.tests.fullstack import base
from neutron.tests.fullstack.resources import environment
from neutron.tests.unit import testlib_api

load_tests = testlib_api.module_load_tests


class TestQuota(base.BaseFullStackTestCase):

    scenarios = [
        ('DbQuotaDriver',
         {'quota_driver': quota_conf.QUOTA_DB_DRIVER_LEGACY}),
        ('DbQuotaNoLockDriver',
         {'quota_driver': quota_conf.QUOTA_DB_DRIVER_NO_LOCK}),
        ('DbQuotaDriverNull',
         {'quota_driver': quota_conf.QUOTA_DB_DRIVER_NULL}),
    ]

    def setUp(self, *args):
        host_descriptions = [environment.HostDescription()]
        env = environment.Environment(environment.EnvironmentDescription(
            quota_driver=self.quota_driver), host_descriptions)
        super().setUp(env)
        self.tenant_id = uuidutils.generate_uuid()

    def test_create_network_and_port(self):
        network = self.safe_client.create_network(self.tenant_id)
        self.safe_client.create_subnet(self.tenant_id, network['id'],
                                       '20.0.0.0/24')
        port = self.safe_client.create_port(self.tenant_id, network['id'])
        port_id = port['id']
        port = self.safe_client.client.list_ports(id=port_id)['ports'][0]
        self.assertEqual(port_id, port['id'])



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import exceptions

from neutron._i18n import _


class CallbackWrongResourceType(exceptions.NeutronException):
    message = _('Callback for %(resource_type)s returned wrong resource type')


class CallbackNotFound(exceptions.NeutronException):
    message = _('Callback for %(resource_type)s not found')


class CallbacksMaxLimitReached(exceptions.NeutronException):
    message = _("Cannot add multiple callbacks for %(resource_type)s")


class NoAgentDbMixinImplemented(exceptions.NeutronException):
    message = _("RPC callbacks mechanism needs the implementation of "
                "AgentDbMixin in the plugin, as so far it's only designed "
                "to work with agents")
